"""
Manages vehicle components at the filesystem level.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from json import JSONDecodeError
from json import dump as json_dump
from json import load as json_load

# from logging import warning as logging_warning
# from sys import exit as sys_exit
from logging import debug as logging_debug
from logging import error as logging_error
from logging import info as logging_info
from os import makedirs as os_makedirs
from os import path as os_path
from os import walk as os_walk
from re import match as re_match
from typing import Any, Union

from ardupilot_methodic_configurator import _
from ardupilot_methodic_configurator.backend_filesystem_json_with_schema import FilesystemJSONWithSchema
from ardupilot_methodic_configurator.backend_filesystem_program_settings import ProgramSettings
from ardupilot_methodic_configurator.data_model_template_overview import TemplateOverview
from ardupilot_methodic_configurator.data_model_vehicle_components_json_schema import VehicleComponentsJsonSchema


class VehicleComponents:
    """Load and save vehicle components configurations from a JSON file."""

    def __init__(self, save_component_to_system_templates: bool = False) -> None:
        self.vehicle_components_fs: FilesystemJSONWithSchema = FilesystemJSONWithSchema(
            "vehicle_components.json", "vehicle_components_schema.json"
        )
        self.save_component_to_system_templates = save_component_to_system_templates

    def load_schema(self) -> dict:
        """
        Load the JSON schema for vehicle components.

        :return: The schema as a dictionary
        """
        return self.vehicle_components_fs.load_schema()

    def load_component_templates(self) -> dict[str, list[dict]]:
        """
        Load component templates from both system and user templates directories.

        User templates take precedence over system templates.

        :return: The merged templates as a dictionary
        """
        # Load system templates (read-only, comes with software)
        system_templates = self._load_system_templates()

        # Load user templates (user-editable)
        user_templates = self._load_user_templates()

        # Merge templates, with user templates taking precedence
        merged_templates = system_templates.copy()
        for component_name, templates in user_templates.items():
            if component_name not in merged_templates:
                merged_templates[component_name] = []

            # Add each user template, replacing system templates with the same name
            for user_template in templates:
                template_name = user_template.get("name")
                if template_name:
                    # Add a flag to mark this as a user template
                    user_template["is_user_modified"] = True

                    # Find index of system template with same name, if any
                    replaced = False
                    for i, template in enumerate(merged_templates.get(component_name, [])):
                        if template.get("name") == template_name:
                            merged_templates[component_name][i] = user_template
                            replaced = True
                            break

                    if not replaced:
                        merged_templates.setdefault(component_name, []).append(user_template)

        return merged_templates

    def _load_system_templates(self) -> dict[str, list[dict]]:
        """
        Load system component templates.

        :return: The system component templates as a dictionary
        """
        templates_filename = "system_vehicle_components_template.json"
        templates_dir = ProgramSettings.get_templates_base_dir()
        filepath = os_path.join(templates_dir, templates_filename)

        templates = {}
        try:
            with open(filepath, encoding="utf-8-sig") as file:
                templates = json_load(file)
        except FileNotFoundError:
            logging_debug(_("System component templates file '%s' not found."), filepath)
        except JSONDecodeError:
            logging_error(_("Error decoding JSON system component templates from file '%s'."), filepath)
        return templates

    def _load_user_templates(self) -> dict[str, list[dict]]:
        """
        Load user component templates.

        :return: The user component templates as a dictionary
        """
        templates_filename = "user_vehicle_components_template.json"
        templates_dir = ProgramSettings.get_templates_base_dir()
        filepath = os_path.join(templates_dir, templates_filename)

        templates = {}
        try:
            with open(filepath, encoding="utf-8-sig") as file:
                templates = json_load(file)
        except FileNotFoundError:
            logging_debug(_("User component templates file '%s' not found."), filepath)
        except JSONDecodeError:
            logging_error(_("Error decoding JSON user component templates from file '%s'."), filepath)
        return templates

    def save_component_templates(self, templates: dict) -> tuple[bool, str]:  # pylint: disable=too-many-branches
        """
        Save component templates.

        For user templates: Only save templates that are user-modified or not present in system templates.
        For system templates: Merge with existing system templates, adding new ones.

        :param templates: The templates to save
        :return: A tuple of (error_occurred, error_message)
        """
        # Load system templates to compare against
        system_templates = self._load_system_templates()

        # Determine which templates need to be saved to user file
        templates_to_save: dict[str, list[dict[str, Any]]] = {}

        if self.save_component_to_system_templates:
            # For system templates, start with existing system templates
            templates_to_save = system_templates.copy()

            # Then add new templates that don't exist yet
            for component_name, component_templates in templates.items():
                if component_name not in templates_to_save:
                    templates_to_save[component_name] = []

                # Create a mapping of existing template names for this component
                existing_template_names = {t.get("name"): True for t in templates_to_save[component_name]}

                for template in component_templates:
                    template_name = template.get("name")
                    if not template_name:
                        continue

                    # Only add if it doesn't exist in system templates yet
                    if template_name not in existing_template_names:
                        template_copy = template.copy()
                        if "is_user_modified" in template_copy:
                            del template_copy["is_user_modified"]
                        templates_to_save[component_name].append(template_copy)
                        existing_template_names[template_name] = True
        else:
            for component_name, component_templates in templates.items():
                templates_to_save[component_name] = []

                # Create a mapping of system template names for this component
                if component_name in system_templates:
                    system_template_names = {t.get("name"): t for t in system_templates.get(component_name, [])}
                else:
                    system_template_names = {}

                for template in component_templates:
                    template_name = template.get("name")
                    if not template_name:
                        continue

                    # If the template is marked as user-modified, save it
                    if template.get("is_user_modified", False):
                        # Remove the flag before saving
                        template_copy = template.copy()
                        if "is_user_modified" in template_copy:
                            del template_copy["is_user_modified"]
                        templates_to_save[component_name].append(template_copy)
                        continue

                    # If the template exists in system templates, check if it's different
                    if template_name in system_template_names:
                        system_template = system_template_names[template_name]

                        # Deep comparison of data section
                        if template.get("data") != system_template.get("data"):
                            # Template is modified from system version
                            template_copy = template.copy()
                            templates_to_save[component_name].append(template_copy)
                    else:
                        # Template doesn't exist in system templates, so it's user-added
                        templates_to_save[component_name].append(template.copy())

                # Remove empty component entries
                if not templates_to_save[component_name]:
                    del templates_to_save[component_name]

        return self.save_component_templates_to_file(templates_to_save)

    def save_component_templates_to_file(self, templates_to_save: dict[str, list[dict[str, Any]]]) -> tuple[bool, str]:
        if self.save_component_to_system_templates:
            # Save to system templates file
            templates_filename = "system_vehicle_components_template.json"
        else:
            # Save to user templates file
            templates_filename = "user_vehicle_components_template.json"
        templates_dir = ProgramSettings.get_templates_base_dir()

        # Create the directory if it doesn't exist
        try:
            os_makedirs(templates_dir, exist_ok=True)
        except Exception as e:  # pylint: disable=broad-exception-caught
            msg = _("Failed to create templates directory '{}': {}").format(templates_dir, str(e))
            logging_error(msg)
            return True, msg

        # Now create the file path and write to it
        filepath = os_path.join(templates_dir, templates_filename)

        try:
            with open(filepath, "w", encoding="utf-8") as file:
                json_dump(templates_to_save, file, indent=4)
            return False, ""  # Success
        except FileNotFoundError:
            msg = _("File not found when writing to '{}': {}").format(filepath, _("Path not found"))
            logging_error(msg)
        except PermissionError:
            msg = _("Permission denied when writing to file '{}'").format(filepath)
            logging_error(msg)
        except OSError as e:
            msg = _("OS error when writing to file '{}': {}").format(filepath, str(e))
            logging_error(msg)
        except Exception as e:  # pylint: disable=broad-exception-caught
            msg = _("Unexpected error saving templates to file '{}': {}").format(filepath, str(e))
            logging_error(msg)
        return True, msg

    def validate_vehicle_components(self, data: dict) -> tuple[bool, str]:
        """
        Validate vehicle components data against the schema.

        :param data: The vehicle components data to validate
        :return: A tuple of (is_valid, error_message)
        """
        return self.vehicle_components_fs.validate_json_against_schema(data)

    def load_vehicle_components_json_data(self, vehicle_dir: str) -> dict[Any, Any]:
        return self.vehicle_components_fs.load_json_data(vehicle_dir)

    def save_vehicle_components_json_data(self, data: dict, vehicle_dir: str) -> tuple[bool, str]:
        """
        Save the vehicle components data to a JSON file.

        :param data: The vehicle components data to save
        :param vehicle_dir: The directory to save the file in
        :return: A tuple of (error_occurred, error_message)
        """
        return self.vehicle_components_fs.save_json_data(data, vehicle_dir)

    def get_fc_fw_type_from_vehicle_components_json(self) -> str:
        data = self.vehicle_components_fs.data
        components = data["Components"] if data and "Components" in data else None
        if components:
            fw_type: str = components.get("Flight Controller", {}).get("Firmware", {}).get("Type", "")
            if fw_type in self.supported_vehicles():
                return fw_type
            error_msg = _("Firmware type {fw_type} in {filename} is not supported")
            logging_error(error_msg.format(fw_type=fw_type, filename=self.vehicle_components_fs.json_filename))
        return ""

    def get_fc_fw_version_from_vehicle_components_json(self) -> str:
        data = self.vehicle_components_fs.data
        components = data["Components"] if data and "Components" in data else None
        if components:
            version_str: str = components.get("Flight Controller", {}).get("Firmware", {}).get("Version", "")
            version_str = version_str.lstrip().split(" ")[0] if version_str else ""
            if re_match(r"^\d+\.\d+\.\d+$", version_str):
                return version_str
            error_msg = _("FW version string {version_str} on {filename} is invalid")
            logging_error(error_msg.format(version_str=version_str, filename=self.vehicle_components_fs.json_filename))
        return ""

    @staticmethod
    def supported_vehicles() -> tuple[str, ...]:
        return ("AP_Periph", "AntennaTracker", "ArduCopter", "ArduPlane", "ArduSub", "Blimp", "Heli", "Rover", "SITL")

    @staticmethod
    def get_vehicle_components_overviews() -> dict[str, TemplateOverview]:
        """
        Finds all subdirectories of the templates base directory containing a "vehicle_components.json" file.

        Creates a dictionary where the keys are
        the subdirectory names (relative to templates base directory) and the
        values are instances of TemplateOverview.

        :return: A dictionary mapping subdirectory paths to TemplateOverview instances.
        """
        vehicle_components_dict = {}
        file_to_find = VehicleComponents().vehicle_components_fs.json_filename
        template_default_dir = ProgramSettings.get_templates_base_dir()
        for root, _dirs, files in os_walk(template_default_dir):
            if file_to_find in files:
                relative_path = os_path.relpath(root, template_default_dir)
                vehicle_components = VehicleComponents()
                comp_data = vehicle_components.load_vehicle_components_json_data(root)
                if comp_data:
                    comp_data = comp_data.get("Components", {})
                    vehicle_components_overview = TemplateOverview(comp_data)
                    vehicle_components_dict[relative_path] = vehicle_components_overview

        return vehicle_components_dict

    @staticmethod
    def get_vehicle_image_filepath(relative_template_path: str) -> str:
        template_default_dir = ProgramSettings.get_templates_base_dir()
        return os_path.join(template_default_dir, relative_template_path, "vehicle.jpg")

    def wipe_component_info(self) -> None:
        """
        Wipe the vehicle components data by clearing all data from the vehicle_components dictionary.

        This resets the internal state without affecting any files.
        Preserves the complete structure of the dictionary including all branches and leaves,
        but sets leaf values to empty values based on their type.
        """
        data = self.vehicle_components_fs.data
        if data is not None:
            self._recursively_clear_dict(data)
            default_data = {
                "RC Receiver": {
                    "FC Connection": {"Type": "RCin/SBUS", "Protocol": "All"},
                },
                "Telemetry": {
                    "FC Connection": {"Type": "SERIAL1", "Protocol": "MAVLink2"},
                },
                "Battery Monitor": {
                    "FC Connection": {"Type": "Analog", "Protocol": "Analog Voltage and Current"},
                },
                "Battery": {
                    "Specifications": {
                        "Chemistry": "Lipo",
                        "Volt per cell max": 4.2,
                        "Volt per cell low": 3.6,
                        "Volt per cell crit": 3.55,
                    },
                },
                "ESC": {
                    "FC Connection": {"Type": "Main Out", "Protocol": "Normal"},
                },
                "Motors": {
                    "Specifications": {"Poles": 14},
                },
                "GNSS Receiver": {
                    "FC Connection": {"Type": "SERIAL3", "Protocol": "AUTO"},
                },
            }

            def merge_defaults(target: dict, defaults: dict) -> None:
                for key, default_value in defaults.items():
                    if key not in target or not target[key]:
                        target[key] = default_value
                    elif isinstance(default_value, dict) and isinstance(target[key], dict):
                        merge_defaults(target[key], default_value)

            merge_defaults(data["Components"], default_data)

    def _recursively_clear_dict(self, data: Union[dict, list, float, bool, str]) -> None:
        """
        Recursively clear leaf values in a nested dictionary while preserving structure.

        :param data: Dictionary to clear
        """
        if not isinstance(data, dict):
            return

        for key, value in data.items():
            if isinstance(value, dict):
                # If it's a dictionary, recurse deeper
                self._recursively_clear_dict(value)
            elif isinstance(value, list):
                # If it's a list, preserve it but empty it
                data[key] = []
            elif isinstance(value, bool):
                data[key] = False
            elif isinstance(value, int):
                data[key] = 0
            elif isinstance(value, float):
                data[key] = 0.0
            elif isinstance(value, str):
                data[key] = ""
            else:
                data[key] = None


def main() -> None:
    """Main function for standalone execution."""
    vehicle_components = VehicleComponents()
    schema = VehicleComponentsJsonSchema(vehicle_components.load_schema())

    component_property_paths = [
        ("Flight Controller", "Product", "Manufacturer"),
        ("Flight Controller", "Product", "Model"),
        ("Flight Controller", "Product", "URL"),
        ("Flight Controller", "Product", "Version"),
        ("Flight Controller", "Firmware", "Type"),
        ("Flight Controller", "Firmware", "Version"),
        ("Flight Controller", "Specifications", "MCU Series"),
        ("Flight Controller", "Notes"),
        ("Flight Controller",),
        ("Flight Controller", "Product"),
        ("Flight Controller", "Firmware"),
        ("Flight Controller", "Specifications"),
        ("Frame",),
        ("Frame", "Product"),
        ("Frame", "Specifications"),
        ("Frame", "Product", "Manufacturer"),
        ("Frame", "Product", "Model"),
        ("Frame", "Product", "URL"),
        ("Frame", "Product", "Version"),
        ("Frame", "Specifications", "TOW min Kg"),
        ("Frame", "Specifications", "TOW max Kg"),
        ("Frame", "Notes"),
    ]

    logging_info("\nTesting description lookup for sample paths:")
    logging_info("=============================================")

    for path in component_property_paths:
        desc, is_optional = schema.get_component_property_description(path)
        msg = f"Path: {path}\nDescription: {desc}, Optional: {is_optional}"
        logging_info(msg)
        logging_info("-" * 50)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
A combobox GUI with support for complex lists.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import contextlib
import tkinter as tk
import tkinter.font as tkfont
from argparse import ArgumentParser, Namespace
from logging import basicConfig as logging_basicConfig
from logging import critical as logging_critical
from logging import debug as logging_debug
from logging import getLevelName as logging_getLevelName
from logging import warning as logging_warning
from platform import system as platform_system
from sys import exit as sys_exit
from tkinter import Label, Toplevel, ttk
from typing import Optional, Union

from ardupilot_methodic_configurator import _
from ardupilot_methodic_configurator.common_arguments import add_common_arguments
from ardupilot_methodic_configurator.frontend_tkinter_autoresize_combobox import update_combobox_width


def setup_combobox_mousewheel_handling(combobox: ttk.Combobox) -> None:
    """
    Set up mouse wheel handling for a combobox to prevent value changes when scrolling.

    This function adds event bindings to prevent the combobox value from changing
    when the user scrolls the mouse wheel over it while the dropdown is closed.
    When the dropdown is closed, the mouse wheel event is propagated to the parent
    widget to allow normal scrolling behavior in the containing widget.

    Args:
        combobox: The ttk.Combobox widget to configure

    """
    # Track dropdown state to control mouse wheel behavior
    combobox.dropdown_is_open = False  # type: ignore[attr-defined]

    def handle_mousewheel(event: tk.Event) -> Optional[str]:
        """Handle mouse wheel events - propagate to parent when dropdown is closed."""
        if not combobox.dropdown_is_open:  # type: ignore[attr-defined]
            # Propagate the wheel event to the parent widget
            combobox.master.event_generate("<MouseWheel>", delta=event.delta)
            return "break"  # Prevent default combobox behavior
        return None  # Allow default behavior when dropdown is open

    def dropdown_opened(_event: Optional[tk.Event] = None) -> None:
        """Mark dropdown as open."""
        combobox.dropdown_is_open = True  # type: ignore[attr-defined]

    def dropdown_closed(_event: Optional[tk.Event] = None) -> None:
        """Mark dropdown as closed."""
        combobox.dropdown_is_open = False  # type: ignore[attr-defined]    # Bind mouse wheel events (Windows and Linux)

    combobox.bind("<MouseWheel>", handle_mousewheel, add="+")
    combobox.bind("<Button-4>", handle_mousewheel, add="+")
    combobox.bind("<Button-5>", handle_mousewheel, add="+")

    # Track dropdown open/close events
    combobox.bind("<<ComboboxDropdown>>", dropdown_opened, add="+")
    combobox.bind("<FocusOut>", dropdown_closed, add="+")


# SPDX-SnippetBegin
# SPDX-License-Identifier: MPL-2.0
# SPDX-SnippetCopyrightText: 2022 geraldew


# https://dev.to/geraldew/python-tkinter-an-exercise-in-wrapping-the-combobox-ndb
class PairTupleCombobox(ttk.Combobox):  # pylint: disable=too-many-ancestors
    """
    A custom Combobox widget that displays a list of key-value tuples.

    This widget processes the list of tuples to separate keys and values for display purposes and allows for the selection
    of a tuple based on its key.
    """

    def __init__(
        self,
        master,  # noqa: ANN001
        list_pair_tuple: list[tuple[str, str]],
        selected_element: Union[None, str],
        cb_name: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.cb_name = cb_name
        self.list_keys: list[str] = []
        self.list_shows: list[str] = []
        self.set_entries_tuple(list_pair_tuple, selected_element)
        self.bind("<Configure>", self.on_combo_configure, add="+")

        # Add keyboard navigation for up/down arrow keys
        self.bind("<Up>", self._on_key_up, add="+")
        self.bind("<Down>", self._on_key_down, add="+")

        # Add focus handling for keyboard events (enable focus reception)
        self.bind("<FocusIn>", lambda _: None, add="+")

        # Apply mouse wheel handling to this combobox instance
        setup_combobox_mousewheel_handling(self)

    def set_entries_tuple(self, list_pair_tuple: list[tuple[str, str]], selected_element: Union[None, str]) -> None:
        # Clear existing entries before setting new ones
        self.list_keys.clear()
        self.list_shows.clear()
        self.append_entries_tuple(list_pair_tuple, selected_element)

    def append_entries_tuple(self, list_pair_tuple: list[tuple[str, str]], selected_element: Union[None, str]) -> None:
        if isinstance(list_pair_tuple, list):
            for tpl in list_pair_tuple:
                self.list_keys.append(tpl[0])
                self.list_shows.append(tpl[1])
        elif isinstance(list_pair_tuple, dict):
            for key, value in list_pair_tuple.items():
                self.list_keys.append(key)
                self.list_shows.append(value)
        else:
            logging_critical(_("list_pair_tuple must be a list of tuples or a dictionary, not %s"), type(list_pair_tuple))
            sys_exit(1)
        self["values"] = tuple(self.list_shows)

        if selected_element:
            default_key_index = self._find_and_validate_selected_element(selected_element)
            if default_key_index is not None:
                self.current(default_key_index)
                update_combobox_width(self)
        else:
            logging_debug(_("No %s combobox element selected"), self.cb_name)

    def _find_and_validate_selected_element(self, selected_element: str) -> Union[int, None]:
        """
        Find the index of the selected element in list_keys.

        Args:
            selected_element: The element to find (should be a list_key)

        Returns:
            The index of the element in list_keys, or None if not found

        """
        try:
            # Find the element in list_keys
            return self.list_keys.index(selected_element)
        except ValueError:
            # Element not found - this should not happen if data model is clean
            logging_critical(
                _("%s combobox selected string '%s' not found in keys %s"),
                self.cb_name,
                selected_element,
                self.list_keys,
            )
            return None

    def get_selected_key(self) -> Union[str, None]:
        try:
            i_index = self.current()
            # self.current() returns -1 if no item is selected
            if i_index < 0:
                return None
            return self.list_keys[i_index]
        except IndexError:
            return None

    # SPDX-SnippetEnd

    def _on_key_up(self, _event: tk.Event) -> str:
        """Handle up arrow key navigation."""
        if not self.list_keys:
            return "break"

        try:
            current_index = self.current()
            # Move up (decrease index), stop at the beginning
            if current_index > 0:
                new_index = current_index - 1
                self.current(new_index)
                # Highlight the text to show selection
                self.selection_range(0, tk.END)
                # Force the combobox to update visually
                self.update_idletasks()
                # Generate a virtual ComboboxSelected event to trigger callbacks
                self.event_generate("<<ComboboxSelected>>")

        except (ValueError, IndexError):
            # Current selection not found or invalid, select first item
            if self.list_keys:
                self.current(0)
                self.update_idletasks()
                self.event_generate("<<ComboboxSelected>>")

        # Prevent the default behavior
        return "break"

    def _on_key_down(self, _event: tk.Event) -> str:
        """Handle down arrow key navigation."""
        if not self.list_keys:
            return "break"

        try:
            current_index = self.current()
            # Move down (increase index), stop at the end
            if current_index < len(self.list_keys) - 1:
                new_index = current_index + 1
                self.current(new_index)
                # Highlight the text to show selection
                self.selection_range(0, tk.END)
                # Force the combobox to update visually
                self.update_idletasks()
                # Generate a virtual ComboboxSelected event to trigger callbacks
                self.event_generate("<<ComboboxSelected>>")

        except (ValueError, IndexError):
            # Current selection not found or invalid, select first item
            if self.list_keys:
                self.current(0)
                # Highlight the text to show selection
                self.selection_range(0, tk.END)
                self.update_idletasks()
                self.event_generate("<<ComboboxSelected>>")

        # Prevent the default behavior
        return "break"

    # https://stackoverflow.com/questions/39915275/change-width-of-dropdown-listbox-of-a-ttk-combobox
    def on_combo_configure(self, event: tk.Event) -> None:
        combo = event.widget
        style = ttk.Style()
        # check if the combobox already has the "postoffset" property
        current_combo_style = combo.cget("style") or "TCombobox"
        if len(style.lookup(current_combo_style, "postoffset")) > 0:
            return
        combo_values = combo.cget("values")
        if len(combo_values) == 0:
            return
        longest_value = max(combo_values, key=len)
        font = tkfont.nametofont("TkDefaultFont")
        width = font.measure(longest_value + "0000") - event.width
        if width < 0:
            # no need to make the popdown smaller
            return
        # create an unique style name using widget's id
        unique_name = f"Combobox{combo.winfo_id()}"
        # the new style must inherit from current widget style (unless it's our custom style!)
        style_name = current_combo_style if unique_name in current_combo_style else f"{unique_name}.{current_combo_style}"

        if isinstance(combo, ttk.Combobox):
            style.configure(style_name, postoffset=(0, 0, width, 0))
            combo.configure(style=style_name)


class PairTupleComboboxTooltip(PairTupleCombobox):  # pylint: disable=too-many-ancestors
    """
    A subclass of PairTupleCombobox that enhances the dropdown experience by displaying tooltips for the highlighted entry.

    This class extends the functionality of PairTupleCombobox by adding tooltips to the items displayed in the dropdown menu.
    When an item is selected or hovered over, a tooltip appears on the right side of the combobox, showing the full value and
      descriptive text of the currently highlighted, potentially truncated, combobox entry.

    The tooltip updates dynamically in real-time based on:
    1. Cursor movement within the dropdown list
    2. Mouse hover over the dropdown items

    Key features:
    - Provides immediate feedback on long entries without scrolling
    - Improves usability by offering more information at a glance
    - Maintains focus on the selected item while providing additional details

    Behavior:
    - Tooltip disappears automatically when:
      a) An item is selected from the dropdown
      b) The dropdown is closed (either by selection or pressing Esc)
    """

    def __init__(
        self,
        master,  # noqa: ANN001
        list_pair_tuple: list[tuple[str, str]],
        selected_element: Union[None, str],
        cb_name: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(master, list_pair_tuple, selected_element, cb_name, *args, **kwargs)
        self.tooltip: Union[None, Toplevel] = None

        # Bind events related to the dropdown
        pd = self.tk.call("ttk::combobox::PopdownWindow", self)
        lb = pd + ".f.l"
        self._bind(("bind", lb), "<KeyRelease>", self.on_key_release, None)  # type: ignore[attr-defined]
        self._bind(("bind", lb), "<Motion>", self.on_motion, None)  # type: ignore[attr-defined]
        self._bind(("bind", lb), "<Escape>", self.on_escape_press, None)  # type: ignore[attr-defined]
        self.bind("<<ComboboxSelected>>", self.on_combobox_selected, None)

    def on_key_release(self, _event: Union[None, tk.Event]) -> None:
        """Get the keyboard highlighted index and create a tooltip for it."""
        pd = self.tk.call("ttk::combobox::PopdownWindow", self)
        lb = pd + ".f.l"
        if self.tk.call(lb, "curselection"):
            highlighted_index = int(self.tk.call(lb, "curselection")[0])
            self.create_tooltip_from_index(highlighted_index)

    def on_motion(self, event: tk.Event) -> None:
        """Get the mouse highlighted index and create a tooltip for it."""
        pd = self.tk.call("ttk::combobox::PopdownWindow", self)
        lb = pd + ".f.l"
        index = self.tk.call(lb, "index", f"@{event.x},{event.y}")
        self.create_tooltip_from_index(int(index))

    def create_tooltip_from_index(self, index: int) -> None:
        with contextlib.suppress(IndexError):
            self.create_tooltip(f"{self.list_keys[index]}: {self.list_shows[index]}")

    def create_tooltip(self, text: str) -> None:
        self.destroy_tooltip()
        try:
            if self.tooltip is None or self.tooltip.winfo_exists():
                self.tooltip = Toplevel(self.master)
                self.tooltip.wm_overrideredirect(boolean=True)
                x = self.winfo_rootx() + 150  # Adjust width as needed
                y = self.winfo_rooty()
                self.tooltip.wm_geometry(f"+{x}+{y}")
                label = Label(self.tooltip, text=text, background="#ffffe0", relief="solid", borderwidth=1, justify=tk.LEFT)
                label.pack(fill="both", expand=True)
        except tk.TclError:
            # If there's no active item, we don't need to update the tooltip
            pass

    def on_combobox_selected(self, _event: Union[None, tk.Event]) -> None:
        self.destroy_tooltip()

    def on_escape_press(self, _event: Union[None, tk.Event]) -> None:
        self.destroy_tooltip()

    def destroy_tooltip(self) -> None:
        if hasattr(self, "tooltip") and self.tooltip and self.tooltip.winfo_exists():
            self.tooltip.destroy()
            self.tooltip = None


def argument_parser() -> Namespace:
    """
    Parses command-line arguments for the script.

    This function sets up an argument parser to handle the command-line arguments for the script.

    Returns:
    argparse.Namespace: An object containing the parsed arguments.

    """
    parser = ArgumentParser(
        description=_(
            "A GUI for testing the PairTupleCombobox. "
            "Not to be used directly, but through the main ArduPilot methodic configurator script."
        )
    )
    return add_common_arguments(parser).parse_args()


def main() -> None:
    argsp = argument_parser()

    logging_basicConfig(level=logging_getLevelName(argsp.loglevel), format="%(asctime)s - %(levelname)s - %(message)s")

    logging_warning(
        _("This main is for testing and development only, usually the PairTupleCombobox is called from another script")
    )

    root = tk.Tk()
    root.title("Random String Selector")
    root.geometry("400x100")  # Set a size for the window

    # Generate 20 random strings between 4 and 70 characters
    import random  # pylint: disable=import-outside-toplevel # noqa: PLC0415
    import string  # pylint: disable=import-outside-toplevel # noqa: PLC0415

    random_strings = [
        "".join(random.choices(string.ascii_letters + string.digits, k=random.randint(4, 70)))  # noqa: S311
        for _ in range(20)
    ]

    # Create the PairTupleCombobox2 instance
    tuple_pairs = [(str(i), random_string) for i, random_string in enumerate(random_strings)]
    combobox = PairTupleCombobox(root, tuple_pairs, None, "Random Strings")

    style = ttk.Style()
    style.theme_use("alt")  # Use an alternative theme for better appearance

    font_size = tkfont.nametofont("TkDefaultFont").cget("size")
    font_size -= 2 if platform_system() == "Windows" else -1
    combobox.config(state="readonly", width=9, font=("TkDefaultFont", font_size))

    # Pack the combobox into the main window
    combobox.pack(pady=10, padx=10)

    # Set the initial value
    combobox.set(str(1))

    tuple_pairs = [("1", "One"), ("2", "Two"), ("3", "Test")]
    # Create a PairTupleComboboxTooltip instance
    tooltip_combobox = PairTupleComboboxTooltip(root, tuple_pairs, None, "Tooltip Items")

    # Place the tooltip below the combobox
    tooltip_combobox.place(relx=0.5, rely=0.7, anchor=tk.CENTER)

    # Start the Tkinter event loop
    root.mainloop()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Tests for the annotate_params.py script.

These are the unit tests for the python script that fetches online ArduPilot
parameter documentation (if not cached) and adds it to the specified file or
to all *.param and *.parm files in the specified directory.

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import contextlib
import os
import tempfile
import unittest
from unittest import mock
from unittest.mock import patch
from xml.etree import ElementTree as ET  # no parsing, just data-structure manipulation

import pytest
import requests  # type: ignore[import-untyped]
from defusedxml import ElementTree as DET  # noqa: N814, just parsing, no data-structure manipulation

from ardupilot_methodic_configurator.annotate_params import (
    BASE_URL,
    PARAM_DEFINITION_XML_FILE,
    create_doc_dict,
    extract_parameter_name_and_validate,
    format_columns,
    get_env_proxies,
    get_xml_data,
    get_xml_url,
    main,
    missionplanner_sort,
    parse_arguments,
    print_read_only_params,
    remove_prefix,
    split_into_lines,
    update_parameter_documentation,
)
from ardupilot_methodic_configurator.data_model_par_dict import Par, ParDict

# pylint: disable=too-many-lines, protected-access


@pytest.fixture
def mock_update() -> mock.Mock:
    with patch("ardupilot_methodic_configurator.annotate_params.update_parameter_documentation") as mock_fun:
        yield mock_fun


@pytest.fixture
def mock_get_xml_dir() -> mock.Mock:
    with patch("ardupilot_methodic_configurator.annotate_params.get_xml_dir") as mock_fun:
        yield mock_fun


@pytest.fixture
def mock_get_xml_url() -> mock.Mock:
    with patch("ardupilot_methodic_configurator.annotate_params.get_xml_url") as mock_fun:
        yield mock_fun


class TestParamDocsUpdate(unittest.TestCase):  # pylint: disable=missing-class-docstring, too-many-public-methods
    def setUp(self) -> None:
        # Create a temporary directory
        self.temp_dir = tempfile.mkdtemp()

        # Create a temporary file
        # pylint: disable=consider-using-with
        self.temp_file = tempfile.NamedTemporaryFile(delete=False)  # noqa: SIM115
        # pylint: enable=consider-using-with

        # Create a dictionary of parameter documentation
        self.doc_dict = {
            "PARAM1": {
                "humanName": "Param 1",
                "documentation": ["Documentation for Param 1"],
                "fields": {"Field1": "Value1", "Field2": "Value2"},
                "values": {"Code1": "Value1", "Code2": "Value2"},
            },
            "PARAM2": {
                "humanName": "Param 2",
                "documentation": ["Documentation for Param 2"],
                "fields": {"Field3": "Value3", "Field4": "Value4"},
                "values": {"Code3": "Value3", "Code4": "Value4"},
            },
            "PARAM_1": {
                "humanName": "Param _ 1",
                "documentation": ["Documentation for Param_1"],
                "fields": {"Field_1": "Value_1", "Field_2": "Value_2"},
                "values": {"Code_1": "Value_1", "Code_2": "Value_2"},
            },
        }

    @patch("builtins.open", new_callable=mock.mock_open, read_data="<root></root>")
    @patch("os.path.isfile")
    @patch("ardupilot_methodic_configurator.data_model_par_dict.ParDict.load_param_file_into_dict")
    def test_get_xml_data_local_file(self, mock_load_param, mock_isfile, mock_open_) -> None:
        # Mock the isfile function to return True
        mock_isfile.return_value = True

        # Mock the load_param_file_into_dict function to raise FileNotFoundError
        mock_load_param.side_effect = FileNotFoundError

        # Call the function with a local file
        result = get_xml_data("/path/to/local/file/", ".", "test.xml", "ArduCopter")

        # Check the result
        assert isinstance(result, ET.Element)

        # Assert that the file was opened correctly
        mock_open_.assert_called_once_with(os.path.join(".", "test.xml"), encoding="utf-8")

    @patch("requests.get")
    def test_get_xml_data_remote_file(self, mock_get) -> None:
        # Mock the response
        mock_get.return_value.status_code = 200
        mock_get.return_value.text = "<root></root>"

        # Remove the test.xml file if it exists
        with contextlib.suppress(FileNotFoundError):
            os.remove("test.xml")

        # Call the function with a remote file
        result = get_xml_data("http://example.com/", ".", "test.xml", "ArduCopter")

        # Check the result
        assert isinstance(result, ET.Element)

        # Assert that the requests.get function was called once
        mock_get.assert_called_once_with("http://example.com/test.xml", timeout=5)

    @patch("os.path.isfile")
    @patch("ardupilot_methodic_configurator.data_model_par_dict.ParDict.load_param_file_into_dict")
    def test_get_xml_data_script_dir_file(self, mock_load_param, mock_isfile) -> None:
        # Mock the isfile function to return False for the current directory and True for the script directory
        def side_effect(_filename) -> bool:
            return True

        mock_isfile.side_effect = side_effect

        # Mock the load_param_file_into_dict function to raise FileNotFoundError
        mock_load_param.side_effect = FileNotFoundError

        # Mock the open function to return a dummy XML string
        mock_open = mock.mock_open(read_data="<root></root>")
        with patch("builtins.open", mock_open):
            # Call the function with a filename that exists in the script directory
            result = get_xml_data(BASE_URL, ".", PARAM_DEFINITION_XML_FILE, "ArduCopter")

        # Check the result
        assert isinstance(result, ET.Element)

        # Assert that the file was opened correctly
        mock_open.assert_called_once_with(os.path.join(".", PARAM_DEFINITION_XML_FILE), encoding="utf-8")

    def test_get_xml_data_no_requests_package(self) -> None:
        # Temporarily remove the requests module
        with patch.dict("sys.modules", {"requests": None}):
            # Remove the test.xml file if it exists
            with contextlib.suppress(FileNotFoundError):
                os.remove("test.xml")

            # Call the function with a remote file
            with pytest.raises(SystemExit):
                get_xml_data("http://example.com/", ".", "test.xml", "ArduCopter")

    @patch("requests.get")
    def test_get_xml_data_request_failure(self, mock_get) -> None:
        # Mock the response
        mock_get.side_effect = requests.exceptions.RequestException

        # Remove the test.xml file if it exists
        with contextlib.suppress(FileNotFoundError):
            os.remove("test.xml")

        # Call the function with a remote file
        with pytest.raises(SystemExit):
            get_xml_data("http://example.com/", ".", "test.xml", "ArduCopter")

    @patch("requests.get")
    def test_get_xml_data_valid_xml(self, mock_get) -> None:
        # Mock the response
        mock_get.return_value.status_code = 200
        mock_get.return_value.text = "<root></root>"

        # Call the function with a remote file
        result = get_xml_data("http://example.com/", ".", "test.xml", "ArduCopter")

        # Check the result
        assert isinstance(result, ET.Element)

    @patch("requests.get")
    def test_get_xml_data_invalid_xml(self, mock_get) -> None:
        # Mock the response
        mock_get.return_value.status_code = 200
        mock_get.return_value.text = "<root><invalid></root>"

        # Remove the test.xml file if it exists
        with contextlib.suppress(FileNotFoundError):
            os.remove("test.xml")

        # Call the function with a remote file
        with pytest.raises(ET.ParseError):
            get_xml_data("http://example.com/", ".", "test.xml", "ArduCopter")

    @patch("requests.get")
    @patch("os.path.isfile")
    def test_get_xml_data_missing_file(self, mock_isfile, mock_get) -> None:
        # Mock the isfile function to return False
        mock_isfile.return_value = False
        # Mock the requests.get call to raise FileNotFoundError
        mock_get.side_effect = FileNotFoundError

        # Remove the test.xml file if it exists
        with contextlib.suppress(FileNotFoundError):
            os.remove("test.xml")

        # Call the function with a local file
        with pytest.raises(FileNotFoundError):
            get_xml_data("/path/to/local/file/", ".", "test.xml", "ArduCopter")

    @patch("requests.get")
    def test_get_xml_data_network_issue(self, mock_get) -> None:
        # Mock the response
        mock_get.side_effect = requests.exceptions.ConnectionError

        # Call the function with a remote file
        with pytest.raises(SystemExit):
            get_xml_data("http://example.com/", ".", "test.xml", "ArduCopter")

    def test_remove_prefix(self) -> None:
        # Test case 1: Normal operation
        assert remove_prefix("prefix_test", "prefix_") == "test"

        # Test case 2: Prefix not present
        assert remove_prefix("test", "prefix_") == "test"

        # Test case 3: Empty string
        assert remove_prefix("", "prefix_") == ""

    def test_split_into_lines(self) -> None:
        # Test case 1: Normal operation
        string_to_split = "This is a test string. It should be split into several lines."
        maximum_line_length = 12
        expected_output = ["This is a", "test string.", "It should be", "split into", "several", "lines."]
        assert split_into_lines(string_to_split, maximum_line_length) == expected_output

        # Test case 2: String shorter than maximum line length
        string_to_split = "Short"
        maximum_line_length = 10
        expected_output = ["Short"]
        assert split_into_lines(string_to_split, maximum_line_length) == expected_output

        # Test case 3: Empty string
        string_to_split = ""
        maximum_line_length = 10
        expected_output = []
        assert split_into_lines(string_to_split, maximum_line_length) == expected_output

    def test_create_doc_dict(self) -> None:
        # Mock XML data
        xml_data = """
        <root>
            <param name="PARAM1" humanName="Param 1" documentation="Documentation for Param 1">
                <field name="Field1">Value1</field>
                <field name="Field2">Value2</field>
                <values>
                    <value code="Code1">Value1</value>
                    <value code="Code2">Value2</value>
                </values>
            </param>
            <param name="PARAM2" humanName="Param 2" documentation="Documentation for Param 2">
                <field name="Units">m/s</field>
                <field name="UnitText">meters per second</field>
                <values>
                    <value code="Code3">Value3</value>
                    <value code="Code4">Value4</value>
                </values>
            </param>
        </root>
        """
        root = DET.fromstring(xml_data)

        # Expected output
        expected_output = {
            "PARAM1": {
                "humanName": "Param 1",
                "documentation": ["Documentation for Param 1"],
                "fields": {"Field1": "Value1", "Field2": "Value2"},
                "values": {"Code1": "Value1", "Code2": "Value2"},
            },
            "PARAM2": {
                "humanName": "Param 2",
                "documentation": ["Documentation for Param 2"],
                "fields": {"Units": "m/s (meters per second)"},
                "values": {"Code3": "Value3", "Code4": "Value4"},
            },
        }

        # Call the function with the mock XML data
        result = create_doc_dict(root, "VehicleType")

        # Check the result
        assert result == expected_output

    def test_format_columns(self) -> None:
        # Define the input
        values = {
            "Key1": "Value1",
            "Key2": "Value2",
            "Key3": "Value3",
            "Key4": "Value4",
            "Key5": "Value5",
            "Key6": "Value6",
            "Key7": "Value7",
            "Key8": "Value8",
            "Key9": "Value9",
            "Key10": "Value10",
            "Key11": "Value11",
            "Key12": "Value12",
        }

        # Define the expected output
        expected_output = [
            "Key1: Value1                                         Key7: Value7",
            "Key2: Value2                                         Key8: Value8",
            "Key3: Value3                                         Key9: Value9",
            "Key4: Value4                                         Key10: Value10",
            "Key5: Value5                                         Key11: Value11",
            "Key6: Value6                                         Key12: Value12",
        ]

        # Call the function with the input
        result = format_columns(values)

        # Check the result
        assert result == expected_output

        assert not format_columns({})

    def test_update_parameter_documentation(self) -> None:
        # Write some initial content to the temporary file
        with open(self.temp_file.name, "w", encoding="utf-8") as file:
            file.write("PARAM1 100\n")

        # Call the function with the temporary file
        update_parameter_documentation(self.doc_dict, self.temp_file.name)

        # Read the updated content from the temporary file
        with open(self.temp_file.name, encoding="utf-8") as file:
            updated_content = file.read()

        # Check if the file has been updated correctly
        assert "Param 1" in updated_content
        assert "Documentation for Param 1" in updated_content
        assert "Field1: Value1" in updated_content
        assert "Field2: Value2" in updated_content
        assert "Code1: Value1" in updated_content
        assert "Code2: Value2" in updated_content

    def test_update_parameter_documentation_sorting_none(self) -> None:
        # Write some initial content to the temporary file
        # With stray leading and trailing whitespaces
        with open(self.temp_file.name, "w", encoding="utf-8") as file:
            file.write("PARAM2 100\n PARAM_1 100 \nPARAM3 3\nPARAM4 4\nPARAM5 5\nPARAM1 100\n")

        # Call the function with the temporary file
        update_parameter_documentation(self.doc_dict, self.temp_file.name)

        # Read the updated content from the temporary file
        with open(self.temp_file.name, encoding="utf-8") as file:
            updated_content = file.read()

        expected_content = """# Param 2
# Documentation for Param 2
# Field3: Value3
# Field4: Value4
# Code3: Value3
# Code4: Value4
PARAM2 100

# Param _ 1
# Documentation for Param_1
# Field_1: Value_1
# Field_2: Value_2
# Code_1: Value_1
# Code_2: Value_2
PARAM_1 100
PARAM3 3
PARAM4 4
PARAM5 5

# Param 1
# Documentation for Param 1
# Field1: Value1
# Field2: Value2
# Code1: Value1
# Code2: Value2
PARAM1 100
"""
        assert updated_content == expected_content

    def test_update_parameter_documentation_sorting_missionplanner(self) -> None:
        # Write some initial content to the temporary file
        with open(self.temp_file.name, "w", encoding="utf-8") as file:
            file.write("PARAM2 100 # ignore, me\nPARAM_1\t100\nPARAM1,100\n")

        # Call the function with the temporary file
        update_parameter_documentation(self.doc_dict, self.temp_file.name, "missionplanner")

        # Read the updated content from the temporary file
        with open(self.temp_file.name, encoding="utf-8") as file:
            updated_content = file.read()

        expected_content = """# Param _ 1
# Documentation for Param_1
# Field_1: Value_1
# Field_2: Value_2
# Code_1: Value_1
# Code_2: Value_2
PARAM_1\t100

# Param 1
# Documentation for Param 1
# Field1: Value1
# Field2: Value2
# Code1: Value1
# Code2: Value2
PARAM1,100

# Param 2
# Documentation for Param 2
# Field3: Value3
# Field4: Value4
# Code3: Value3
# Code4: Value4
PARAM2 100 # ignore, me
"""
        assert updated_content == expected_content

    def test_update_parameter_documentation_sorting_mavproxy(self) -> None:
        # Write some initial content to the temporary file
        with open(self.temp_file.name, "w", encoding="utf-8") as file:
            file.write("PARAM2 100\nPARAM_1\t100\nPARAM1,100\n")

        # Call the function with the temporary file
        update_parameter_documentation(self.doc_dict, self.temp_file.name, "mavproxy")

        # Read the updated content from the temporary file
        with open(self.temp_file.name, encoding="utf-8") as file:
            updated_content = file.read()

        expected_content = """# Param 1
# Documentation for Param 1
# Field1: Value1
# Field2: Value2
# Code1: Value1
# Code2: Value2
PARAM1,100

# Param 2
# Documentation for Param 2
# Field3: Value3
# Field4: Value4
# Code3: Value3
# Code4: Value4
PARAM2 100

# Param _ 1
# Documentation for Param_1
# Field_1: Value_1
# Field_2: Value_2
# Code_1: Value_1
# Code_2: Value_2
PARAM_1\t100
"""
        assert updated_content == expected_content

    def test_update_parameter_documentation_invalid_line_format(self) -> None:
        # Write some initial content to the temporary file with an invalid line format
        with open(self.temp_file.name, "w", encoding="utf-8") as file:
            file.write("%INVALID_LINE_FORMAT\n")

        # Call the function with the temporary file
        with pytest.raises(SystemExit) as cm:
            update_parameter_documentation(self.doc_dict, self.temp_file.name)

        # Check if the SystemExit exception contains the expected message
        assert cm.value.code == "Invalid line in input file"

    @patch("logging.Logger.info")
    def test_print_read_only_params(self, mock_info) -> None:
        # Mock XML data
        xml_data = """
        <root>
            <param name="PARAM1" humanName="Param 1" documentation="Documentation for Param 1">
                <field name="ReadOnly">True</field>
                <field name="Field1">Value1</field>
                <field name="Field2">Value2</field>
                <values>
                    <value code="Code1">Value1</value>
                    <value code="Code2">Value2</value>
                </values>
            </param>
            <param name="PARAM2" humanName="Param 2" documentation="Documentation for Param 2">
                <field name="Field3">Value3</field>
                <field name="Field4">Value4</field>
                <values>
                    <value code="Code3">Value3</value>
                    <value code="Code4">Value4</value>
                </values>
            </param>
        </root>
        """
        root = DET.fromstring(xml_data)
        doc_dict = create_doc_dict(root, "VehicleType")

        # Call the function with the mock XML data
        print_read_only_params(doc_dict)

        # Check if the parameter name was logged
        mock_info.assert_has_calls([mock.call("ReadOnly parameters:"), mock.call("PARAM1")])

    def test_update_parameter_documentation_invalid_target(self) -> None:
        with pytest.raises(ValueError, match=r"Target 'invalid_target' is neither a file nor a directory\."):
            update_parameter_documentation(self.doc_dict, "invalid_target")

    def test_invalid_parameter_name(self) -> None:
        # Write some initial content to the temporary file
        with open(self.temp_file.name, "w", encoding="utf-8") as file:
            file.write("INVALID_$PARAM 100\n")

        # Call the function with the temporary file
        with pytest.raises(SystemExit):
            update_parameter_documentation(self.doc_dict, self.temp_file.name)

    def test_update_parameter_documentation_too_long_parameter_name(self) -> None:
        # Write some initial content to the temporary file
        with open(self.temp_file.name, "w", encoding="utf-8") as file:
            file.write("TOO_LONG_PARAMETER_NAME 100\n")

        # Call the function with the temporary file
        with pytest.raises(SystemExit):
            update_parameter_documentation(self.doc_dict, self.temp_file.name)

    @patch("logging.Logger.warning")
    def test_missing_parameter_documentation(self, mock_warning) -> None:
        # Write some initial content to the temporary file
        with open(self.temp_file.name, "w", encoding="utf-8") as file:
            file.write("MISSING_DOC_PARA 100\n")

        # Call the function with the temporary file
        update_parameter_documentation(self.doc_dict, self.temp_file.name)

        # Check if the warnings were logged
        mock_warning.assert_has_calls(
            [
                mock.call("Read file %s with %d parameters, but only %s of which got documented", self.temp_file.name, 1, 0),
                mock.call("No documentation found for: %s", "MISSING_DOC_PARA"),
            ]
        )

    def test_empty_parameter_file(self) -> None:
        # Call the function with the temporary file
        update_parameter_documentation(self.doc_dict, self.temp_file.name)

        # Read the updated content from the temporary file
        with open(self.temp_file.name, encoding="utf-8") as file:
            updated_content = file.read()

        # Check if the file is still empty
        assert updated_content == ""

    def test_get_xml_url_valid_vehicles(self) -> None:
        """Test get_xml_url with all valid vehicle types."""
        vehicle_types = ["ArduCopter", "ArduPlane", "Rover", "ArduSub", "AntennaTracker", "AP_Periph", "Blimp", "Heli", "SITL"]
        for vehicle in vehicle_types:
            url = get_xml_url(vehicle, "4.3")
            assert url.startswith(BASE_URL)
            assert "stable-4.3" in url
            assert url.endswith("/")

    def test_get_xml_url_invalid_vehicle(self) -> None:
        """Test get_xml_url with invalid vehicle type."""
        with pytest.raises(ValueError, match=r"Vehicle type 'InvalidVehicle' is not supported\."):
            get_xml_url("InvalidVehicle", "4.3")

    def test_split_into_lines_edge_cases(self) -> None:
        """Test split_into_lines with edge cases."""
        # Test with various line lengths
        # Function will return largest possible chunks based on max length
        assert split_into_lines("a b c", 2) == ["a", "b", "c"]
        assert split_into_lines("", 10) == []

    def test_format_columns_edge_cases(self) -> None:
        """Test format_columns with edge cases."""
        # Empty dictionary
        assert not format_columns({})

        # Single item
        assert format_columns({"Key": "Value"}) == ["Key: Value"]

        # Test with different max widths
        values = {"K1": "V1", "K2": "V2"}
        assert len(format_columns(values, max_width=20)[0]) <= 20

        # Test with many columns
        many_values = {f"Key{i}": f"Value{i}" for i in range(20)}
        result = format_columns(many_values, max_width=200, max_columns=5)
        assert all(len(line) <= 200 for line in result)

    def test_create_doc_dict_edge_cases(self) -> None:
        """Test create_doc_dict with edge cases."""
        # Test with empty XML
        empty_root = ET.Element("root")
        assert not create_doc_dict(empty_root, "ArduCopter")

        # Test with missing attributes
        param = ET.SubElement(empty_root, "param")
        assert not create_doc_dict(empty_root, "ArduCopter")

        # Test with minimal valid param
        param.set("name", "TEST_PARAM")
        param.set("humanName", "Test Parameter")
        param.set("documentation", "Test documentation")
        doc_dict = create_doc_dict(empty_root, "ArduCopter")
        assert "TEST_PARAM" in doc_dict
        assert doc_dict["TEST_PARAM"]["humanName"] == "Test Parameter"

    @patch("os.path.isfile")
    def test_update_parameter_documentation_sorting(self, mock_isfile) -> None:
        """Test parameter sorting in update_parameter_documentation."""
        # Mock file existence check
        mock_isfile.return_value = True

        test_content = "PARAM_Z 100\nPARAM_A 200\nPARAM_M 300\n"

        # Create a real temporary file for testing
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_file:
            temp_file.write(test_content)
            temp_file_name = temp_file.name

        doc = {
            "PARAM_Z": {"humanName": "Z", "documentation": ["Z doc"], "fields": {}, "values": {}},
            "PARAM_A": {"humanName": "A", "documentation": ["A doc"], "fields": {}, "values": {}},
            "PARAM_M": {"humanName": "M", "documentation": ["M doc"], "fields": {}, "values": {}},
        }

        try:
            # Test MissionPlanner sorting
            update_parameter_documentation(doc, temp_file_name, "missionplanner")

            with open(temp_file_name, encoding="utf-8") as f:
                content = f.read()

            # Verify content and order
            assert "PARAM_A" in content
            assert "PARAM_M" in content
            assert "PARAM_Z" in content
            assert content.index("PARAM_A") < content.index("PARAM_M") < content.index("PARAM_Z")

            # Test MAVProxy sorting
            # Reset file content
            with open(temp_file_name, "w", encoding="utf-8") as f:
                f.write(test_content)

            update_parameter_documentation(doc, temp_file_name, "mavproxy")

            with open(temp_file_name, encoding="utf-8") as f:
                content = f.read()

            # Verify content for MAVProxy format
            assert "PARAM_A" in content
            assert "PARAM_M" in content
            assert "PARAM_Z" in content

        finally:
            # Clean up
            os.unlink(temp_file_name)

    def test_extract_parameter_name_and_validate_invalid_cases(self) -> None:
        """Test parameter name validation with invalid cases."""
        # Test invalid parameter name pattern
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("invalid_param 100", "test.param", 1)

        # Test too long parameter name
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("VERY_LONG_PARAMETER_NAME_THAT_EXCEEDS_LIMIT 100", "test.param", 1)

        # Test invalid separator
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("PARAM:100", "test.param", 1)

    def test_par_class_methods(self) -> None:
        """Test Par class methods."""
        # Test equality
        par1 = Par(100.0, "comment1")
        par2 = Par(100.0, "comment1")
        par3 = Par(200.0, "comment2")

        assert par1 == par2
        assert par1 != par3
        assert par1 != "not a Par object"

        # Test load_param_file with invalid values
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as tf:
            tf.write("PARAM1 invalid_value\n")
            tf.flush()

            with pytest.raises(SystemExit):
                ParDict.load_param_file_into_dict(tf.name)

    def test_format_params_methods(self) -> None:
        """Test ParDict._format_params method."""
        param_dict = ParDict({"PARAM1": Par(100.0, "comment1"), "PARAM2": Par(200.0), "PARAM3": Par(300.0)})

        # Test MissionPlanner format
        mp_format = param_dict._format_params("missionplanner")
        assert any("PARAM1,100" in line for line in mp_format)
        assert any("# comment1" in line for line in mp_format)

        # Test MAVProxy format
        mavproxy_format = param_dict._format_params("mavproxy")
        # Use correct spacing format - 16 chars for name, 8 for value
        assert any("PARAM1           100.000000" in line for line in mavproxy_format)
        assert any("# comment1" in line for line in mavproxy_format)

        # Test invalid format
        with pytest.raises(SystemExit):
            param_dict._format_params("invalid_format")


class AnnotateParamsTest(unittest.TestCase):
    """Test annotate parameters."""

    def test_arg_parser_valid_arguments(self) -> None:
        test_args = ["annotate_params", "--vehicle-type", "ArduCopter", "--sort", "none", "parameters"]
        with patch("sys.argv", test_args):
            args = parse_arguments()
            assert args.vehicle_type == "ArduCopter"
            assert args.sort == "none"
            assert args.target == "parameters"
            assert args.verbose is False
            assert args.max_line_length == 100

    def test_arg_parser_invalid_vehicle_type(self) -> None:
        test_args = ["annotate_params", "--vehicle-type", "InvalidType", "--sort", "none", "parameters"]
        with patch("sys.argv", test_args), pytest.raises(SystemExit):
            parse_arguments()

    def test_arg_parser_invalid_sort_option(self) -> None:
        test_args = ["annotate_params", "--vehicle-type", "ArduCopter", "--sort", "invalid", "parameters"]
        with patch("sys.argv", test_args), pytest.raises(SystemExit):
            parse_arguments()

    def test_arg_parser_invalid_line_length_option(self) -> None:
        test_args = ["annotate_params", "--vehicle-type", "ArduCopter", "--sort", "none", "-m", "invalid", "parameters"]
        with patch("sys.argv", test_args), pytest.raises(SystemExit):
            parse_arguments()


class TestAnnotateParamsExceptionHandling(unittest.TestCase):
    """Test parameter exception handling."""

    @pytest.mark.usefixtures("mock_update", "mock_get_xml_dir", "mock_get_xml_url")
    @patch("builtins.open", new_callable=mock.mock_open)
    def test_main_ioerror(self, mock_file) -> None:
        with patch("ardupilot_methodic_configurator.annotate_params.parse_arguments") as mock_arg_parser:
            mock_arg_parser.return_value = mock.Mock(
                vehicle_type="ArduCopter",
                firmware_version="4.0",
                target=".",
                sort="none",
                delete_documentation_annotations=False,
                verbose=False,
            )
            mock_file.side_effect = OSError("Mocked IO Error")

            with pytest.raises(SystemExit) as cm:
                main()

            assert cm.value.code in [1, 2]

    @pytest.mark.usefixtures("mock_update", "mock_get_xml_dir", "mock_get_xml_url")
    @patch("builtins.open", new_callable=mock.mock_open)
    def test_main_oserror(self, mock_file) -> None:
        with patch("ardupilot_methodic_configurator.annotate_params.parse_arguments") as mock_arg_parser:
            mock_arg_parser.return_value = mock.Mock(
                vehicle_type="ArduCopter",
                firmware_version="4.0",
                target=".",
                sort="none",
                delete_documentation_annotations=False,
                verbose=False,
            )
            mock_file.side_effect = OSError("Mocked OS Error")

            with pytest.raises(SystemExit) as cm:
                main()

            assert cm.value.code in [1, 2]

    @patch("ardupilot_methodic_configurator.annotate_params.get_xml_url")
    def test_get_xml_url_exception(self, mock_get_xml_url_) -> None:
        mock_get_xml_url_.side_effect = ValueError("Mocked Value Error")
        with pytest.raises(ValueError, match=r"Vehicle type 'NonExistingVehicle' is not supported\."):  # noqa: PT012
            get_xml_url("NonExistingVehicle", "4.0")

            @patch("requests.get")
            def test_get_xml_data_remote_file(mock_get) -> None:
                """Test fetching XML data from remote file."""
                # Mock the response
                mock_get.return_value.status_code = 200
                mock_get.return_value.text = "<root></root>"

                # Remove the test.xml file if it exists
                with contextlib.suppress(FileNotFoundError):
                    os.remove("test.xml")

                # Call the function with a remote file
                result = get_xml_data("http://example.com/", ".", "test.xml", "ArduCopter")

                # Check the result
                assert isinstance(result, ET.Element)

                # Assert that requests.get was called once with correct parameters including proxies
                mock_get.assert_called_once_with("http://example.com/test.xml", timeout=5, proxies=None)

            @patch("requests.get")
            def test_get_xml_data_remote_file_with_proxies(mock_get) -> None:
                """Test fetching XML data with proxy configuration."""
                # Mock environment variables
                with patch.dict(
                    os.environ,
                    {"HTTP_PROXY": "http://proxy:8080", "HTTPS_PROXY": "https://proxy:8080", "NO_PROXY": "localhost"},
                ):
                    # Mock the response
                    mock_get.return_value.status_code = 200
                    mock_get.return_value.text = "<root></root>"

                    # Call the function
                    result = get_xml_data("http://example.com/", ".", "test.xml", "ArduCopter")

                    # Check the result
                    assert isinstance(result, ET.Element)

                    # Assert that requests.get was called with proxy settings
                    expected_proxies = {"http": "http://proxy:8080", "https": "https://proxy:8080", "no_proxy": "localhost"}
                    mock_get.assert_called_once_with("http://example.com/test.xml", timeout=5, proxies=expected_proxies)

            @patch("requests.get")
            def test_get_xml_data_remote_file_no_proxies(mock_get) -> None:
                """Test fetching XML data with no proxy configuration."""
                # Clear environment variables
                with patch.dict(os.environ, {}, clear=True):
                    # Mock the response
                    mock_get.return_value.status_code = 200
                    mock_get.return_value.text = "<root></root>"

                    # Call the function
                    result = get_xml_data("http://example.com/", ".", "test.xml", "ArduCopter")

                    # Check the result
                    assert isinstance(result, ET.Element)

                    # Assert that requests.get was called with no proxies
                    mock_get.assert_called_once_with("http://example.com/test.xml", timeout=5, proxies=None)

    @patch.dict(
        "os.environ",
        {
            "HTTP_PROXY": "http://proxy-server:8080",
            "HTTPS_PROXY": "https://proxy-server:8080",
            "NO_PROXY": "localhost,127.0.0.1",
        },
    )
    def test_get_env_proxies_with_proxies(self) -> None:
        """Test getting proxies from environment variables."""
        proxies = get_env_proxies()
        assert proxies is not None
        assert proxies["http"] == "http://proxy-server:8080"
        assert proxies["https"] == "https://proxy-server:8080"
        assert proxies["no_proxy"] == "localhost,127.0.0.1"

    @patch.dict("os.environ", {}, clear=True)
    def test_get_env_proxies_without_proxies(self) -> None:
        """Test getting proxies when environment variables are not set."""
        proxies = get_env_proxies()
        assert proxies is None

    @patch.dict("os.environ", {"http_proxy": "http://lowercase-proxy:8080"})
    def test_get_env_proxies_lowercase(self) -> None:
        """Test getting proxies from lowercase environment variables."""
        proxies = get_env_proxies()
        assert proxies is not None
        assert proxies["http"] == "http://lowercase-proxy:8080"

    def test_extract_parameter_name_and_validate_edge_cases(self) -> None:
        """Test extract_parameter_name_and_validate with various edge cases."""
        # Valid parameter names with different separators
        assert extract_parameter_name_and_validate("PARAM1,100", "test.param", 1) == "PARAM1"
        assert extract_parameter_name_and_validate("PARAM2 100", "test.param", 1) == "PARAM2"
        assert extract_parameter_name_and_validate("PARAM3\t100", "test.param", 1) == "PARAM3"

        # Parameter name must start with capital letter
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("param4 100", "test.param", 1)

        # Parameter name can't have special characters
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("PARAM-5 100", "test.param", 1)

        # Parameter must be followed by a valid separator
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("PARAM6:100", "test.param", 1)

    def test_par_class_methods_comprehensive(self) -> None:
        """Test Par class methods more comprehensively."""
        # Test Par.__init__ and equality
        par1 = Par(10.5, "comment")
        par2 = Par(10.5, "comment")
        par3 = Par(10.5, "different comment")
        par4 = Par(11.0, "comment")

        assert par1 == par2
        assert par1 != par3
        assert par1 != par4
        assert par1 != "not a Par object"

        # Test load_param_file_into_dict with various formats
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as tf:
            tf.write("PARAM1,10.5 # comment1\n")
            tf.write("PARAM2 20.75 # comment2\n")
            tf.write("PARAM3\t30.25\n")  # tab separator, no comment
            tf.write("# This is a full line comment\n")
            tf.write("\n")  # empty line
            tf_name = tf.name

        param_dict = ParDict.load_param_file_into_dict(tf_name)
        os.unlink(tf_name)

        assert len(param_dict) == 3
        assert param_dict["PARAM1"].value == 10.5
        assert param_dict["PARAM1"].comment == "comment1"
        assert param_dict["PARAM2"].value == 20.75
        assert param_dict["PARAM2"].comment == "comment2"
        assert param_dict["PARAM3"].value == 30.25
        assert param_dict["PARAM3"].comment is None

        # Test load_param_file_into_dict with duplicate parameters
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as tf:
            tf.write("PARAM1,10.5\n")
            tf.write("PARAM1,20.5\n")  # Duplicate parameter
            tf_name = tf.name

        with pytest.raises(SystemExit):
            ParDict.load_param_file_into_dict(tf_name)
        os.unlink(tf_name)

    def test_missionplanner_sort_function(self) -> None:
        """Test the missionplanner_sort function."""
        # Basic sorting
        params = ["Z_PARAM", "A_PARAM", "C_PARAM", "B_PARAM"]
        sorted_params = sorted(params, key=missionplanner_sort)
        assert sorted_params == ["A_PARAM", "B_PARAM", "C_PARAM", "Z_PARAM"]

        # Complex sorting with underscores
        params = ["RC_FEEL", "RC_MAP_ROLL", "RC_MAP_PITCH", "RC_1_MIN"]
        sorted_params = sorted(params, key=missionplanner_sort)
        assert sorted_params == ["RC_1_MIN", "RC_FEEL", "RC_MAP_PITCH", "RC_MAP_ROLL"]

        # Sort with line content
        lines = ["RC_FEEL 100", "RC_MAP_ROLL 200", "RC_MAP_PITCH 300", "RC_1_MIN 400"]
        sorted_lines = sorted(lines, key=missionplanner_sort)
        assert sorted_lines == ["RC_1_MIN 400", "RC_FEEL 100", "RC_MAP_PITCH 300", "RC_MAP_ROLL 200"]

    @patch("logging.Logger.warning")
    def test_update_parameter_documentation_edge_cases(self, mock_warning) -> None:  # pylint: disable=unused-argument
        """Test update_parameter_documentation with edge cases."""
        # Create a simplified doc_dict for testing
        doc_dict = {
            "PARAM1": {
                "humanName": "Parameter 1",
                "documentation": ["Documentation for Parameter 1"],
                "fields": {},
                "values": {},
            },
            "PARAM2": {
                "humanName": "Parameter 2",
                "documentation": ["Documentation for Parameter 2"],
                "fields": {},
                "values": {},
            },
        }

        # Create a temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create an empty file
            empty_file = os.path.join(temp_dir, "empty.param")
            with open(empty_file, "w", encoding="utf-8") as f:
                pass

            # Create file with only comments and whitespace
            comments_file = os.path.join(temp_dir, "comments.param")
            with open(comments_file, "w", encoding="utf-8") as f:
                f.write("# This is a comment\n\n# Another comment\n")

            # Create file with comments and parameters
            mixed_file = os.path.join(temp_dir, "mixed.param")
            with open(mixed_file, "w", encoding="utf-8") as f:
                f.write("# A comment\nPARAM1,100\n\n# Another comment\nPARAM2 200\n")

            # Test with empty file
            update_parameter_documentation(doc_dict, empty_file)
            with open(empty_file, encoding="utf-8") as f:
                content = f.read()
            assert content == ""

            # Test with comments-only file
            update_parameter_documentation(doc_dict, comments_file)
            with open(comments_file, encoding="utf-8") as f:
                content = f.read()
            assert content == ""

            # Test with mixed content and delete_documentation_annotations=True
            update_parameter_documentation(doc_dict, mixed_file, delete_documentation_annotations=True)
            with open(mixed_file, encoding="utf-8") as f:
                content = f.read()
            assert "# Parameter 1" not in content
            assert "PARAM1,100" in content
            assert "PARAM2 200" in content

            # Write the file again with annotations
            with open(mixed_file, "w", encoding="utf-8") as f:
                f.write("# A comment\nPARAM1,100\n\n# Another comment\nPARAM2 200\n")

            # Test with mixed content and regular annotations
            update_parameter_documentation(doc_dict, mixed_file)
            with open(mixed_file, encoding="utf-8") as f:
                content = f.read()
            assert "# Parameter 1" in content
            assert "# Documentation for Parameter 1" in content
            assert "PARAM1,100" in content
            assert "# Parameter 2" in content
            assert "# Documentation for Parameter 2" in content
            assert "PARAM2 200" in content

    def test_extract_parameter_name_and_validate_comprehensive(self) -> None:
        """Comprehensive testing of parameter name extraction and validation."""
        # Valid parameter names with different separators
        assert extract_parameter_name_and_validate("PARAM1,100", "test.param", 1) == "PARAM1"
        assert extract_parameter_name_and_validate("PARAM2 100", "test.param", 1) == "PARAM2"
        assert extract_parameter_name_and_validate("PARAM3\t100", "test.param", 1) == "PARAM3"
        assert extract_parameter_name_and_validate("A123_XYZ,100", "test.param", 1) == "A123_XYZ"

        # Parameter name with trailing whitespace before separator
        assert extract_parameter_name_and_validate("PARAM4   \t100", "test.param", 1) == "PARAM4"

        # Invalid parameter format (not starting with uppercase letter)
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("param5 100", "test.param", 1)

        # Invalid parameter format (invalid character)
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("PARAM-5 100", "test.param", 1)

        # Invalid separator
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("PARAM6:100", "test.param", 1)

        # Parameter name too long
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("ABCDEFGHIJKLMNOPQR 100", "test.param", 1)

        # Empty line
        with pytest.raises(SystemExit):
            extract_parameter_name_and_validate("", "test.param", 1)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

"""
Tests for the frontend_tkinter_show.py file.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from collections.abc import Generator
from unittest.mock import MagicMock, patch

import pytest

from ardupilot_methodic_configurator.frontend_tkinter_show import (
    Tooltip,
    show_error_message,
    show_no_connection_error,
    show_no_param_files_error,
    show_tooltip,
)

# pylint: disable=redefined-outer-name


# Fixtures
@pytest.fixture
def mock_tk() -> Generator[MagicMock, None, None]:
    with patch("tkinter.Tk") as mock:
        mock.return_value.withdraw.return_value = None
        mock.return_value.destroy.return_value = None
        yield mock


@pytest.fixture
def mock_widget() -> MagicMock:
    widget = MagicMock()
    widget.winfo_rootx.return_value = 100
    widget.winfo_rooty.return_value = 200
    widget.winfo_width.return_value = 50
    widget.winfo_height.return_value = 30
    return widget


@pytest.fixture
def mock_toplevel() -> Generator[tuple[MagicMock, MagicMock], None, None]:
    with patch("tkinter.Toplevel") as mock:
        mock_instance = MagicMock()
        mock.return_value = mock_instance
        yield mock, mock_instance


@pytest.fixture
def mock_label() -> Generator[MagicMock, None, None]:
    with patch("tkinter.ttk.Label") as mock:
        mock.return_value.pack.return_value = None
        yield mock


# Tests for show_error_message
@pytest.mark.parametrize(
    ("title", "message"),
    [
        ("Test Title", "Test Message"),
        ("Test & Title", "Test\nMessage with & special < chars >"),
    ],
)
def test_show_error_message(mock_tk, title, message) -> None:
    with patch("tkinter.messagebox.showerror") as mock_showerror, patch("tkinter.ttk.Style"):
        # Call the function with test parameters
        show_error_message(title, message)

        # Assert that the Tkinter Tk class was instantiated
        mock_tk.assert_called_once()

        # Assert that the Tkinter messagebox.showerror function was called with the correct parameters
        mock_showerror.assert_called_once_with(title, message)

        # Assert that the Tkinter Tk instance's withdraw method was called
        mock_tk.return_value.withdraw.assert_called_once()

        # Assert that the Tkinter Tk instance's destroy method was called
        mock_tk.return_value.destroy.assert_called_once()


# Tests for show_no_param_files_error
def test_show_no_param_files_error(mock_tk) -> None:
    with patch("tkinter.messagebox.showerror") as mock_showerror, patch("tkinter.ttk.Style"):
        show_no_param_files_error("test_dir")

        mock_tk.assert_called_once()
        mock_showerror.assert_called_once_with(
            "No Parameter Files Found",
            (
                "No intermediate parameter files found in the selected 'test_dir' vehicle directory.\n"
                "Please select and step inside a vehicle directory containing valid ArduPilot intermediate parameter files."
                "\n\nMake sure to step inside the directory (double-click) and not just select it."
            ),
        )
        mock_tk.return_value.withdraw.assert_called_once()
        mock_tk.return_value.destroy.assert_called_once()


# Tests for show_no_connection_error
def test_show_no_connection_error(mock_tk) -> None:
    with patch("tkinter.messagebox.showerror") as mock_showerror, patch("tkinter.ttk.Style"):
        show_no_connection_error("test_error")

        mock_tk.assert_called_once()
        mock_showerror.assert_called_once_with(
            "No Connection to the Flight Controller",
            "test_error\n\nPlease connect a flight controller to the PC,\nwait at least 7 seconds and retry.",
        )
        mock_tk.return_value.withdraw.assert_called_once()
        mock_tk.return_value.destroy.assert_called_once()


# Tests for show_tooltip function
def test_show_tooltip(mock_widget) -> None:
    # Mock the Tooltip class to avoid creating actual tkinter windows
    with patch("ardupilot_methodic_configurator.frontend_tkinter_show.Tooltip") as mock_tooltip_class:
        mock_tooltip_instance = MagicMock()
        mock_tooltip_class.return_value = mock_tooltip_instance

        # Call the function with test parameters
        tooltip = show_tooltip(mock_widget, "Test Tooltip Message")

        # Check that tooltip object was created and is the mocked instance
        assert tooltip is mock_tooltip_instance

        # Verify Tooltip was called with correct parameters (position_below=True is the default)
        mock_tooltip_class.assert_called_once_with(mock_widget, "Test Tooltip Message", position_below=True)


# Tests for Tooltip class
def test_tooltip_init(mock_widget) -> None:
    # Test non-macOS initialization (where tooltip is created immediately)
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel") as mock_toplevel,
    ):
        mock_instance = MagicMock()
        mock_toplevel.return_value = mock_instance

        tooltip = Tooltip(mock_widget, "Test tooltip text")

        # Verify attribute initialization
        assert tooltip.widget == mock_widget
        assert tooltip.text == "Test tooltip text"
        assert tooltip.tooltip is not None  # On non-macOS, tooltip is created immediately
        assert tooltip.position_below is True

        # Verify event bindings for non-macOS
        mock_widget.bind.assert_any_call("<Enter>", tooltip.show)
        mock_widget.bind.assert_any_call("<Leave>", tooltip.hide)

        # Verify that Toplevel was created and configured
        mock_toplevel.assert_called_once_with(mock_widget)
        mock_instance.wm_overrideredirect.assert_called_once_with(boolean=True)


def test_tooltip_init_macos(mock_widget) -> None:
    # Test macOS initialization (where tooltip is not created immediately)
    with patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Darwin"):
        tooltip = Tooltip(mock_widget, "Test tooltip text")

        # Verify attribute initialization
        assert tooltip.widget == mock_widget
        assert tooltip.text == "Test tooltip text"
        assert tooltip.tooltip is None  # On macOS, tooltip is not created immediately
        assert tooltip.position_below is True

        # Verify event bindings for macOS
        mock_widget.bind.assert_any_call("<Enter>", tooltip.create_show)
        mock_widget.bind.assert_any_call("<Leave>", tooltip.destroy_hide)


def test_tooltip_show(mock_widget, mock_toplevel, mock_label) -> None:
    _, mock_toplevel_instance = mock_toplevel

    tooltip = Tooltip(mock_widget, "Test tooltip")
    tooltip.show()

    # Verify toplevel window was created
    assert tooltip.tooltip is not None

    # Calculate expected positioning
    expected_x = mock_widget.winfo_rootx() + min(mock_widget.winfo_width() // 2, 100)
    expected_y = mock_widget.winfo_rooty() + mock_widget.winfo_height()

    # Check positioning
    mock_toplevel_instance.geometry.assert_called_with(f"+{expected_x}+{expected_y}")

    # Verify label was created with right text
    mock_label.assert_called_once()
    assert mock_label.call_args[1]["text"] == "Test tooltip"


def test_tooltip_show_position_above(mock_widget, mock_toplevel) -> None:
    _, mock_toplevel_instance = mock_toplevel

    # Create tooltip positioned above widget
    tooltip = Tooltip(mock_widget, "Test tooltip", position_below=False)
    tooltip.show()

    # Calculate expected positioning (should be above the widget)
    expected_x = mock_widget.winfo_rootx() + min(mock_widget.winfo_width() // 2, 100)
    expected_y = mock_widget.winfo_rooty() - 10  # positioned above

    # Check positioning
    mock_toplevel_instance.geometry.assert_called_with(f"+{expected_x}+{expected_y}")

    # Create tooltip positioned above widget


def test_tooltip_hide(mock_widget, mock_toplevel) -> None:
    _, mock_toplevel_instance = mock_toplevel

    tooltip = Tooltip(mock_widget, "Test tooltip")
    tooltip.show()

    # Verify tooltip created
    assert tooltip.tooltip is not None

    # Clear the call count from initialization
    mock_toplevel_instance.withdraw.reset_mock()

    # Now hide the tooltip
    tooltip.hide()

    # On non-macOS systems, hide() should call withdraw() but not destroy the tooltip
    mock_toplevel_instance.withdraw.assert_called_once()
    # The tooltip instance should still exist
    assert tooltip.tooltip is not None
    # destroy should not have been called
    mock_toplevel_instance.destroy.assert_not_called()


def test_tooltip_destroy_hide_macos(mock_widget) -> None:
    # Test macOS specific destroy behavior
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Darwin"),
        patch("tkinter.Toplevel") as mock_toplevel,
    ):
        mock_instance = MagicMock()
        mock_toplevel.return_value = mock_instance

        tooltip = Tooltip(mock_widget, "Test tooltip")

        # Simulate the tooltip being shown (created on macOS)
        tooltip.create_show()

        # Verify tooltip created
        assert tooltip.tooltip is not None

        # Now destroy_hide the tooltip (macOS behavior)
        tooltip.destroy_hide()

        # On macOS systems, destroy_hide() should destroy the tooltip
        mock_instance.destroy.assert_called_once()
        # The tooltip instance should be None
        assert tooltip.tooltip is None


def test_tooltip_show_hide_event_handling(mock_widget) -> None:
    """Test that tooltip properly handles Enter and Leave events."""
    # Test non-macOS event handling
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel") as mock_toplevel,
    ):
        mock_toplevel_instance = MagicMock()
        mock_toplevel.return_value = mock_toplevel_instance

        # Create tooltip instance
        tooltip = Tooltip(mock_widget, "Test tooltip")

        # Verify the widget was bound to the correct events
        mock_widget.bind.assert_any_call("<Enter>", tooltip.show)
        mock_widget.bind.assert_any_call("<Leave>", tooltip.hide)

        # Test show method with event
        mock_event = MagicMock()
        tooltip.show(mock_event)

        # Verify tooltip is visible after show
        mock_toplevel_instance.deiconify.assert_called_once()

        # Reset mock for hide test
        mock_toplevel_instance.reset_mock()

        # Test hide method with event
        tooltip.hide(mock_event)

        # Verify tooltip is hidden after hide
        mock_toplevel_instance.withdraw.assert_called_once()


def test_tooltip_macos_event_handling(mock_widget) -> None:
    """Test that tooltip properly handles Enter and Leave events on macOS."""
    with patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Darwin"):
        # Create tooltip instance
        tooltip = Tooltip(mock_widget, "Test tooltip")

        # Verify the widget was bound to the correct events for macOS
        mock_widget.bind.assert_any_call("<Enter>", tooltip.create_show)
        mock_widget.bind.assert_any_call("<Leave>", tooltip.destroy_hide)

        # Verify tooltip is initially None on macOS
        assert tooltip.tooltip is None


def test_tooltip_position_tooltip_method(mock_widget) -> None:
    """Test the position_tooltip method directly."""
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel") as mock_toplevel,
    ):
        mock_toplevel_instance = MagicMock()
        mock_toplevel.return_value = mock_toplevel_instance

        # Test positioning below (default)
        tooltip = Tooltip(mock_widget, "Test tooltip")
        tooltip.position_tooltip()

        expected_x = mock_widget.winfo_rootx() + min(mock_widget.winfo_width() // 2, 100)
        expected_y = mock_widget.winfo_rooty() + mock_widget.winfo_height()
        mock_toplevel_instance.geometry.assert_called_with(f"+{expected_x}+{expected_y}")

        # Test positioning above
        tooltip_above = Tooltip(mock_widget, "Test tooltip", position_below=False)
        tooltip_above.position_tooltip()

        expected_y_above = mock_widget.winfo_rooty() - 10
        mock_toplevel_instance.geometry.assert_called_with(f"+{expected_x}+{expected_y_above}")


def test_tooltip_none_tooltip_edge_cases(mock_widget) -> None:
    """Test tooltip behavior when tooltip is None."""
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel") as mock_toplevel,
    ):
        mock_toplevel_instance = MagicMock()
        mock_toplevel.return_value = mock_toplevel_instance

        tooltip = Tooltip(mock_widget, "Test tooltip")

        # Set tooltip to None and test methods don't crash
        tooltip.tooltip = None

        # These should not raise exceptions
        tooltip.show()
        tooltip.hide()
        tooltip.position_tooltip()


def test_tooltip_show_tooltip_function_with_different_positions(mock_widget) -> None:
    """Test the show_tooltip function with different position parameters."""
    with patch("ardupilot_methodic_configurator.frontend_tkinter_show.Tooltip") as mock_tooltip_class:
        mock_tooltip_instance = MagicMock()
        mock_tooltip_class.return_value = mock_tooltip_instance

        # Test with position_below=True (default)
        tooltip1 = show_tooltip(mock_widget, "Test message 1")
        mock_tooltip_class.assert_called_with(mock_widget, "Test message 1", position_below=True)
        assert tooltip1 is mock_tooltip_instance

        # Reset mock for next test
        mock_tooltip_class.reset_mock()

        # Test with position_below=False
        tooltip2 = show_tooltip(mock_widget, "Test message 2", position_below=False)
        mock_tooltip_class.assert_called_with(mock_widget, "Test message 2", position_below=False)
        assert tooltip2 is mock_tooltip_instance


def test_tooltip_create_show_destroy_hide_macos_complete_cycle(mock_widget) -> None:
    """Test complete create/destroy cycle on macOS."""
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Darwin"),
        patch("tkinter.Toplevel") as mock_toplevel,
        patch("tkinter.ttk.Label") as mock_label,
    ):
        mock_toplevel_instance = MagicMock()
        mock_toplevel.return_value = mock_toplevel_instance

        tooltip = Tooltip(mock_widget, "Test tooltip")

        # Initially tooltip should be None on macOS
        assert tooltip.tooltip is None

        # Create and show tooltip
        tooltip.create_show()
        assert tooltip.tooltip is not None
        mock_toplevel.assert_called_once_with(mock_widget)
        mock_label.assert_called_once()

        # Destroy and hide tooltip
        tooltip.destroy_hide()
        assert tooltip.tooltip is None
        mock_toplevel_instance.destroy.assert_called_once()


def test_error_message_functions_comprehensive() -> None:
    """Test all error message functions comprehensively."""
    with (
        patch("tkinter.Tk") as mock_tk,
        patch("tkinter.messagebox.showerror") as mock_showerror,
        patch("tkinter.ttk.Style") as mock_style,  # noqa: F841 # pylint: disable=unused-variable
    ):
        mock_tk_instance = MagicMock()
        mock_tk.return_value = mock_tk_instance

        # Test show_error_message with special characters
        show_error_message("Test & Title", "Test\nMessage with & special < chars >")
        mock_tk.assert_called_once()
        mock_showerror.assert_called_once_with("Test & Title", "Test\nMessage with & special < chars >")
        mock_tk_instance.withdraw.assert_called_once()
        mock_tk_instance.destroy.assert_called_once()

        # Reset mocks
        mock_tk.reset_mock()
        mock_showerror.reset_mock()
        mock_tk_instance.reset_mock()

        # Test show_no_param_files_error
        show_no_param_files_error("test_directory")
        mock_tk.assert_called_once()
        expected_message = (
            "No intermediate parameter files found in the selected 'test_directory' vehicle directory.\n"
            "Please select and step inside a vehicle directory containing valid ArduPilot intermediate parameter files."
            "\n\nMake sure to step inside the directory (double-click) and not just select it."
        )
        mock_showerror.assert_called_once_with("No Parameter Files Found", expected_message)

        # Reset mocks
        mock_tk.reset_mock()
        mock_showerror.reset_mock()
        mock_tk_instance.reset_mock()

        # Test show_no_connection_error
        show_no_connection_error("Connection failed: timeout")
        mock_tk.assert_called_once()
        expected_message = (
            "Connection failed: timeout\n\nPlease connect a flight controller to the PC,\nwait at least 7 seconds and retry."
        )
        mock_showerror.assert_called_once_with("No Connection to the Flight Controller", expected_message)


def test_tooltip_darwin_handling(mock_widget) -> None:
    # Test macOS specific code paths
    with patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Darwin"):
        # Test the successful path
        with patch("tkinter.Toplevel") as mock_toplevel_class:
            # Set up the mock instance
            mock_toplevel_instance = MagicMock()
            mock_toplevel_class.return_value = mock_toplevel_instance

            # Create a tooltip and show it
            tooltip = Tooltip(mock_widget, "Test tooltip")
            tooltip.create_show()

            # Check that tk.call was called (don't check exact parameters)
            assert mock_toplevel_instance.tk.call.called

        # Test the AttributeError exception path with a separate patch
        with patch("tkinter.Toplevel") as mock_toplevel_class:
            # Set up the mock instance with exception
            mock_toplevel_instance = MagicMock()
            mock_toplevel_class.return_value = mock_toplevel_instance

            # We need to set side_effect only for the specific tk.call we're testing
            # First mock with a function that checks arguments
            def side_effect_function(*args) -> None:
                # Only raise AttributeError for the MacWindowStyle call
                if args and args[0] == "::tk::unsupported::MacWindowStyle":
                    raise AttributeError
                # For all other calls, return None or a default value

            mock_toplevel_instance.tk.call.side_effect = side_effect_function

            # Create a new tooltip and show it
            tooltip2 = Tooltip(mock_widget, "Test tooltip")
            tooltip2.create_show()

            # Check the fallback attributes were set
            mock_toplevel_instance.wm_attributes.assert_any_call("-alpha", 1.0)
            mock_toplevel_instance.wm_attributes.assert_any_call("-topmost", True)  # noqa: FBT003


def test_tooltip_non_darwin_handling(mock_widget) -> None:
    # Test non-macOS code path
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel") as mock_toplevel,
    ):
        mock_instance = MagicMock()
        mock_toplevel.return_value = mock_instance

        tooltip = Tooltip(mock_widget, "Test tooltip")
        tooltip.show()

        # Verify overrideredirect was called
        mock_instance.wm_overrideredirect.assert_called_with(boolean=True)


def test_tooltip_edge_case_empty_text(mock_widget) -> None:
    """Test tooltip with empty text."""
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel"),
    ):
        tooltip = Tooltip(mock_widget, "")
        assert tooltip.text == ""


def test_tooltip_edge_case_very_long_text(mock_widget) -> None:
    """Test tooltip with very long text."""
    long_text = "This is a very long tooltip text " * 100
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel"),
    ):
        tooltip = Tooltip(mock_widget, long_text)
        assert tooltip.text == long_text


def test_tooltip_edge_case_special_characters(mock_widget) -> None:
    """Test tooltip with special characters and unicode."""
    special_text = 'Test with ñ, é, ü, ™, ©, €, 🚁, <>&"'
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel"),
    ):
        tooltip = Tooltip(mock_widget, special_text)
        assert tooltip.text == special_text


def test_tooltip_position_calculation_edge_cases(mock_widget) -> None:
    """Test tooltip positioning with edge case widget dimensions."""
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel") as mock_toplevel,
    ):
        mock_toplevel_instance = MagicMock()
        mock_toplevel.return_value = mock_toplevel_instance

        # Test with very small widget
        mock_widget.winfo_width.return_value = 1
        mock_widget.winfo_height.return_value = 1
        mock_widget.winfo_rootx.return_value = 0
        mock_widget.winfo_rooty.return_value = 0

        tooltip = Tooltip(mock_widget, "Test")
        tooltip.position_tooltip()

        # Width is 1, so min(1//2, 100) = min(0, 100) = 0
        expected_x = 0 + 0  # rootx + min(width//2, 100)
        expected_y = 0 + 1  # rooty + height (position_below=True)
        mock_toplevel_instance.geometry.assert_called_with(f"+{expected_x}+{expected_y}")

        # Test with very large widget
        mock_widget.winfo_width.return_value = 1000
        mock_widget.winfo_height.return_value = 500
        mock_widget.winfo_rootx.return_value = 100
        mock_widget.winfo_rooty.return_value = 200

        tooltip.position_tooltip()

        # Width is 1000, so min(1000//2, 100) = min(500, 100) = 100
        expected_x = 100 + 100  # rootx + min(width//2, 100)
        expected_y = 200 + 500  # rooty + height
        mock_toplevel_instance.geometry.assert_called_with(f"+{expected_x}+{expected_y}")


def test_tooltip_macos_exception_handling_comprehensive(mock_widget) -> None:
    """Test comprehensive exception handling on macOS."""
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Darwin"),
        patch("tkinter.Toplevel") as mock_toplevel_class,
    ):
        mock_toplevel_instance = MagicMock()
        mock_toplevel_class.return_value = mock_toplevel_instance

        # Test different types of exceptions that might occur
        def tk_call_side_effect(*args) -> None:
            if args and args[0] == "::tk::unsupported::MacWindowStyle":
                msg = "MacWindowStyle not supported"
                raise AttributeError(msg)

        mock_toplevel_instance.tk.call.side_effect = tk_call_side_effect

        tooltip = Tooltip(mock_widget, "Test tooltip")
        tooltip.create_show()

        # Verify fallback attributes were set
        mock_toplevel_instance.wm_attributes.assert_any_call("-alpha", 1.0)
        mock_toplevel_instance.wm_attributes.assert_any_call("-topmost", True)  # noqa: FBT003
        mock_toplevel_instance.configure.assert_called_with(bg="#ffffe0")


def test_error_functions_with_extreme_inputs() -> None:
    """Test error functions with extreme inputs."""
    with (
        patch("tkinter.Tk") as mock_tk,
        patch("tkinter.messagebox.showerror") as mock_showerror,
        patch("tkinter.ttk.Style"),
    ):
        mock_tk_instance = MagicMock()
        mock_tk.return_value = mock_tk_instance

        # Test with empty strings
        show_error_message("", "")
        mock_showerror.assert_called_with("", "")

        # Test with very long strings
        long_title = "A" * 1000
        long_message = "B" * 10000
        show_error_message(long_title, long_message)
        mock_showerror.assert_called_with(long_title, long_message)

        # Test with None-like empty directory
        show_no_param_files_error("")
        assert mock_showerror.called

        # Test with special path characters
        show_no_param_files_error("C:\\Program Files\\Test & Dir\\")
        assert mock_showerror.called


def test_tooltip_multiple_instances(mock_widget) -> None:
    """Test creating multiple tooltip instances."""
    with (
        patch("ardupilot_methodic_configurator.frontend_tkinter_show.platform_system", return_value="Windows"),
        patch("tkinter.Toplevel"),
    ):
        # Create multiple tooltips for the same widget
        tooltip1 = Tooltip(mock_widget, "Tooltip 1")
        tooltip2 = Tooltip(mock_widget, "Tooltip 2")
        tooltip3 = Tooltip(mock_widget, "Tooltip 3", position_below=False)

        assert tooltip1.text == "Tooltip 1"
        assert tooltip2.text == "Tooltip 2"
        assert tooltip3.text == "Tooltip 3"
        assert tooltip1.position_below is True
        assert tooltip2.position_below is True
        assert tooltip3.position_below is False


def test_show_tooltip_function_stress_test(mock_widget) -> None:
    """Stress test the show_tooltip function."""
    with patch("ardupilot_methodic_configurator.frontend_tkinter_show.Tooltip") as mock_tooltip_class:
        mock_instances = []

        # Create multiple tooltips rapidly
        for i in range(10):
            mock_instance = MagicMock()
            mock_instances.append(mock_instance)
            mock_tooltip_class.return_value = mock_instance

            tooltip = show_tooltip(mock_widget, f"Tooltip {i}", position_below=i % 2 == 0)
            assert tooltip is mock_instance

            # Verify correct call
            expected_position = i % 2 == 0
            mock_tooltip_class.assert_called_with(mock_widget, f"Tooltip {i}", position_below=expected_position)
            mock_tooltip_class.reset_mock()



#!/usr/bin/env python3

"""
Unit tests for the ArduPilotParameter domain model.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from typing import Any

import pytest

from ardupilot_methodic_configurator.data_model_ardupilot_parameter import (
    ArduPilotParameter,
    ParameterUnchangedError,
)
from ardupilot_methodic_configurator.data_model_par_dict import Par

# pylint: disable=redefined-outer-name, protected-access


@pytest.fixture
def param_fixture() -> dict[str, Any]:  # pylint: disable=too-many-locals
    """Create parameter fixtures for testing."""
    # Create a basic parameter
    param_name = "TEST_PARAM"
    param_value = 10.0
    param_comment = "Test comment"
    par_obj = Par(param_value, param_comment)

    # Create parameter with default value
    default_value = 5.0
    default_par = Par(default_value, "Default comment")

    # Create FC value
    fc_value = 15.0

    # Metadata with bitmask, values dict, and constraints
    metadata: dict[str, Any] = {
        "doc_tooltip": "Test tooltip",
        "unit": "m/s",
        "unit_tooltip": "Meters per second",
        "min": 0.0,
        "max": 20.0,
        "values": {"0": "Zero", "10": "Ten", "20": "Twenty"},
        "Bitmask": {0: "Option 1", 1: "Option 2", 2: "Option 3"},
        "Calibration": False,
        "ReadOnly": False,
    }

    # Basic parameter without extras
    basic_param = ArduPilotParameter(param_name, par_obj)

    # Parameter with all attributes
    full_param = ArduPilotParameter(param_name, par_obj, metadata, default_par, fc_value)

    # Forced parameter
    forced_param = ArduPilotParameter(
        param_name, par_obj, metadata, default_par, fc_value, forced_par=Par(param_value, param_comment)
    )

    # Derived parameter
    derived_param = ArduPilotParameter(
        param_name, par_obj, metadata, default_par, fc_value, derived_par=Par(param_value, param_comment)
    )

    # Readonly parameter
    readonly_metadata = metadata.copy()
    readonly_metadata["ReadOnly"] = True
    readonly_param = ArduPilotParameter(param_name, par_obj, readonly_metadata, default_par, fc_value)

    # Calibration parameter
    calibration_metadata = metadata.copy()
    calibration_metadata["Calibration"] = True
    calibration_param = ArduPilotParameter(param_name, par_obj, calibration_metadata, default_par, fc_value)

    return {
        "param_name": param_name,
        "param_value": param_value,
        "param_comment": param_comment,
        "par_obj": par_obj,
        "default_value": default_value,
        "default_par": default_par,
        "fc_value": fc_value,
        "metadata": metadata,
        "basic_param": basic_param,
        "full_param": full_param,
        "forced_param": forced_param,
        "derived_param": derived_param,
        "readonly_param": readonly_param,
        "calibration_param": calibration_param,
    }


def test_basic_properties(param_fixture) -> None:
    """
    Parameter stores and retrieves basic attributes correctly.

    GIVEN: A parameter is created with name, value, comment, default, and FC value
    WHEN: The parameter attributes are accessed
    THEN: All attributes should match the provided values
    """
    # Assert: All basic properties are correctly stored
    assert param_fixture["full_param"].name == param_fixture["param_name"]
    assert param_fixture["full_param"]._new_value == param_fixture["param_value"]
    assert param_fixture["full_param"].change_reason == param_fixture["param_comment"]
    assert param_fixture["full_param"]._default_value == param_fixture["default_value"]
    assert param_fixture["full_param"]._fc_value == param_fixture["fc_value"]
    assert not param_fixture["full_param"].is_forced
    assert not param_fixture["full_param"].is_derived


def test_parameter_type_properties(param_fixture) -> None:
    """
    Parameter correctly identifies its type based on metadata.

    GIVEN: Parameters are created with different metadata configurations
    WHEN: Type properties are checked (readonly, calibration, bitmask)
    THEN: Each parameter should correctly identify its type
    """
    # Assert: Readonly parameter is identified correctly
    assert param_fixture["readonly_param"].is_readonly
    assert not param_fixture["full_param"].is_readonly

    # Assert: Calibration parameter is identified correctly
    assert param_fixture["calibration_param"].is_calibration
    assert not param_fixture["full_param"].is_calibration

    # Assert: Bitmask parameter is identified correctly
    assert param_fixture["full_param"].is_bitmask

    # Arrange: Create a parameter without bitmask metadata
    no_bitmask_metadata = param_fixture["metadata"].copy()
    no_bitmask_metadata.pop("Bitmask")
    no_bitmask_param = ArduPilotParameter(param_fixture["param_name"], param_fixture["par_obj"], no_bitmask_metadata)

    # Assert: Non-bitmask parameter is identified correctly
    assert not no_bitmask_param.is_bitmask


def test_value_comparisons(param_fixture) -> None:
    """
    Parameter correctly compares values between file, default, and flight controller.

    GIVEN: Parameters with different value configurations
    WHEN: Value comparison properties are accessed
    THEN: Comparisons should correctly identify matching or differing values
    """
    # Assert: Parameter with non-default value is identified correctly
    assert not param_fixture["full_param"].new_value_equals_default_value

    # Arrange: Create parameter with value matching default
    default_value_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(param_fixture["default_value"], ""),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["fc_value"],
    )

    # Assert: Parameter with default value is identified correctly
    assert default_value_param.new_value_equals_default_value

    # Assert: FC value presence is detected correctly
    assert param_fixture["full_param"].has_fc_value
    assert not param_fixture["basic_param"].has_fc_value

    # Assert: Difference from FC value is detected correctly
    assert param_fixture["full_param"].is_different_from_fc

    # Arrange: Create parameter matching FC value
    same_as_fc_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(param_fixture["fc_value"], ""),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["fc_value"],
    )

    # Assert: Parameter matching FC value is identified correctly
    assert not same_as_fc_param.is_different_from_fc


def test_metadata_properties(param_fixture) -> None:
    """
    Parameter exposes metadata documentation and constraints correctly.

    GIVEN: Parameters with complete or missing metadata
    WHEN: Metadata properties are accessed
    THEN: Properties should return correct values or appropriate defaults
    """
    # Assert: Parameter with full metadata returns correct values
    assert param_fixture["full_param"].tooltip_new_value == param_fixture["metadata"]["doc_tooltip"]
    assert param_fixture["full_param"].unit == param_fixture["metadata"]["unit"]
    assert param_fixture["full_param"].tooltip_unit == param_fixture["metadata"]["unit_tooltip"]
    assert param_fixture["full_param"].min_value == float(param_fixture["metadata"]["min"])
    assert param_fixture["full_param"].max_value == float(param_fixture["metadata"]["max"])

    # Assert: Parameter without metadata returns appropriate defaults
    assert param_fixture["basic_param"].tooltip_new_value == "No documentation available in apm.pdef.xml for this parameter"
    assert param_fixture["basic_param"].unit == ""
    assert param_fixture["basic_param"].tooltip_unit == "No documentation available in apm.pdef.xml for this parameter"
    assert param_fixture["basic_param"].min_value is None
    assert param_fixture["basic_param"].max_value is None


def test_string_representations(param_fixture) -> None:
    """
    Parameter formats numeric values as strings correctly.

    GIVEN: Parameters with various numeric values
    WHEN: String representation properties are accessed
    THEN: Values should be formatted correctly with trailing zeros removed
    """
    # Assert: Parameter value formatted correctly
    assert param_fixture["full_param"].value_as_string == "10"
    assert param_fixture["full_param"].fc_value_as_string == "15"

    # Assert: Missing FC value shows appropriate message
    assert param_fixture["basic_param"].fc_value_as_string == "N/A"

    # Arrange: Create parameter with decimal value
    decimal_param = ArduPilotParameter(param_fixture["param_name"], Par(10.5, ""))

    # Assert: Decimal value formatted correctly
    assert decimal_param.value_as_string == "10.5"


def test_value_dictionary_operations(param_fixture) -> None:
    """
    Parameter looks up human-readable labels from values dictionary.

    GIVEN: A parameter with a values dictionary mapping numbers to labels
    WHEN: Value lookup operations are performed
    THEN: Correct labels should be returned or None for missing values
    """
    # Assert: Value in dictionary is found and labeled correctly
    assert param_fixture["full_param"].is_in_values_dict
    assert param_fixture["full_param"].get_selected_value_from_dict() == "Ten"

    # Arrange: Create parameter with value not in dictionary
    not_in_dict_param = ArduPilotParameter(param_fixture["param_name"], Par(11.0, ""), param_fixture["metadata"])

    # Assert: Value not in dictionary returns None
    assert not not_in_dict_param.is_in_values_dict
    assert not_in_dict_param.get_selected_value_from_dict() is None

    # Assert: Parameter without values dictionary returns None
    assert not param_fixture["basic_param"].is_in_values_dict
    assert param_fixture["basic_param"].get_selected_value_from_dict() is None


def test_set_new_value(param_fixture) -> None:
    """
    User can update parameter values with validation.

    GIVEN: A regular, forced, or derived parameter
    WHEN: User attempts to set a new value
    THEN: Regular parameters accept changes, forced/derived parameters reject changes
    AND: Setting the same value raises ParameterUnchangedError
    """
    # Arrange: Prepare new value for testing
    new_value = "12.0"

    # Act: User changes value on regular parameter
    result = param_fixture["basic_param"].set_new_value(new_value)

    # Assert: Value is updated correctly
    assert result == 12.0
    assert param_fixture["basic_param"]._new_value == 12.0

    # Act & Assert: User attempts to set same value again
    with pytest.raises(ParameterUnchangedError):
        param_fixture["basic_param"].set_new_value(new_value)

    # Arrange: Store original forced parameter value
    original_value = param_fixture["forced_param"]._new_value

    # Act & Assert: User attempts to change forced parameter
    with pytest.raises(ValueError, match="forced or derived"):
        param_fixture["forced_param"].set_new_value(new_value)

    # Assert: Forced parameter value unchanged
    assert param_fixture["forced_param"]._new_value == original_value

    # Arrange: Store original derived parameter value
    original_value = param_fixture["derived_param"]._new_value

    # Act & Assert: User attempts to change derived parameter
    with pytest.raises(ValueError, match="forced or derived"):
        param_fixture["derived_param"].set_new_value(new_value)

    # Assert: Derived parameter value unchanged
    assert param_fixture["derived_param"]._new_value == original_value


def test_set_change_reason(param_fixture) -> None:
    """
    User can update parameter change reasons with appropriate restrictions.

    GIVEN: Regular, forced, and derived parameters
    WHEN: User attempts to set a change reason
    THEN: Regular parameters accept changes, forced/derived parameters reject changes
    AND: Setting the same reason returns False
    """
    # Arrange: Prepare new comment
    new_comment = "New comment"

    # Act: User sets change reason on regular parameter
    result = param_fixture["full_param"].set_change_reason(new_comment)

    # Assert: Change reason is updated
    assert result
    assert param_fixture["full_param"].change_reason == new_comment

    # Act: User sets same comment again
    result = param_fixture["full_param"].set_change_reason(new_comment)

    # Assert: No change detected
    assert not result

    # Arrange: Create parameter with None comment
    none_comment_param = ArduPilotParameter(param_fixture["param_name"], Par(10.0, None))

    # Act: User sets empty string on parameter with None comment
    result = none_comment_param.set_change_reason("")

    # Assert: Empty string equivalent to None, no change
    assert not result

    # Arrange: Store original forced parameter comment
    original_comment = param_fixture["forced_param"].change_reason

    # Act: User attempts to change forced parameter comment
    result = param_fixture["forced_param"].set_change_reason(new_comment)

    # Assert: Forced parameter comment unchanged
    assert not result
    assert param_fixture["forced_param"].change_reason == original_comment

    # Arrange: Store original derived parameter comment
    original_comment = param_fixture["derived_param"].change_reason

    # Act: User attempts to change derived parameter comment
    result = param_fixture["derived_param"].set_change_reason(new_comment)

    # Assert: Derived parameter comment unchanged
    assert not result
    assert param_fixture["derived_param"].change_reason == original_comment


def test_is_editable(param_fixture) -> None:
    """
    Parameter correctly identifies whether user can edit it.

    GIVEN: Regular, forced, derived, and readonly parameters
    WHEN: Editability is checked
    THEN: Only regular parameters should be editable
    """
    # Assert: Regular parameter is editable
    assert param_fixture["full_param"].is_editable

    # Assert: Forced parameter is not editable
    assert not param_fixture["forced_param"].is_editable

    # Assert: Derived parameter is not editable
    assert not param_fixture["derived_param"].is_editable

    # Assert: Readonly parameter is not editable
    assert not param_fixture["readonly_param"].is_editable


def test_is_dirty(param_fixture) -> None:
    """
    Parameter correctly detects unsaved changes for save workflow.

    GIVEN: Parameters with various value and comment configurations
    WHEN: The is_dirty property is checked
    THEN: Only parameters with changes from file should be marked dirty
    AND: Forced/derived parameters with different computed values should be dirty
    """
    # Arrange: Create parameter with value matching file
    same_value_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(param_fixture["param_value"], param_fixture["param_comment"]),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["param_value"],
    )

    # Assert: Parameter with same value as file is not dirty
    assert not same_value_param.is_dirty

    # Arrange: Create parameter with different FC value but same file value
    different_fc_value_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(param_fixture["param_value"], param_fixture["param_comment"]),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["fc_value"],
    )

    # Assert: FC value difference alone doesn't make parameter dirty
    assert not different_fc_value_param.is_dirty

    # Arrange: Create parameter and change its value
    changed_value_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(param_fixture["param_value"], param_fixture["param_comment"]),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["param_value"],
    )

    # Act: User changes the value
    changed_value_param.set_new_value(str(param_fixture["fc_value"]))

    # Assert: Parameter with user-changed value is dirty
    assert changed_value_param.is_dirty

    # Arrange: Create derived parameter with different computed comment
    derived_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(param_fixture["param_value"], param_fixture["param_comment"]),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["param_value"],
        derived_par=Par(param_fixture["param_value"], "computed comment"),
    )

    # Assert: Derived parameter with different comment is dirty
    assert derived_param.is_dirty

    # Arrange: Create forced parameter with different computed comment
    forced_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(param_fixture["param_value"], param_fixture["param_comment"]),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["param_value"],
        forced_par=Par(param_fixture["param_value"], "forced comment"),
    )

    # Assert: Forced parameter with different comment is dirty
    assert forced_param.is_dirty

    # Assert: Fixture forced parameter with matching values is not dirty
    assert not param_fixture["forced_param"].is_dirty

    # Assert: Fixture derived parameter with matching values is not dirty
    assert not param_fixture["derived_param"].is_dirty


def test_set_forced_or_derived_value(param_fixture) -> None:
    """
    System can update forced/derived parameter values after creation.

    GIVEN: Forced and derived parameters that need value updates
    WHEN: set_forced_or_derived_value is called
    THEN: Values should be updated for forced/derived parameters
    AND: Regular parameters should raise ValueError
    """
    # Arrange: Get forced parameter and prepare new value
    forced_param = param_fixture["forced_param"]
    new_value = 999.0

    # Act: System updates forced parameter value
    forced_param.set_forced_or_derived_value(new_value)

    # Assert: Forced parameter value is updated
    assert forced_param.get_new_value() == new_value

    # Arrange: Get derived parameter and prepare new value
    derived_param = param_fixture["derived_param"]
    new_value = 888.0

    # Act: System updates derived parameter value
    derived_param.set_forced_or_derived_value(new_value)

    # Assert: Derived parameter value is updated
    assert derived_param.get_new_value() == new_value

    # Arrange: Create regular parameter
    regular_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(param_fixture["param_value"], param_fixture["param_comment"]),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["fc_value"],
    )

    # Act & Assert: System attempting to use method on regular parameter fails
    with pytest.raises(ValueError, match="This method is only for forced or derived parameters"):
        regular_param.set_forced_or_derived_value(123.0)


def test_set_forced_or_derived_change_reason(param_fixture) -> None:
    """
    System can update forced/derived parameter change reasons after creation.

    GIVEN: Forced and derived parameters that need reason updates
    WHEN: set_forced_or_derived_change_reason is called
    THEN: Reasons should be updated for forced/derived parameters
    AND: Regular parameters should raise ValueError
    """
    # Arrange: Get forced parameter and prepare new reason
    forced_param = param_fixture["forced_param"]
    new_reason = "New forced reason"

    # Act: System updates forced parameter change reason
    forced_param.set_forced_or_derived_change_reason(new_reason)

    # Assert: Forced parameter change reason is updated
    assert forced_param.change_reason == new_reason

    # Arrange: Get derived parameter and prepare new reason
    derived_param = param_fixture["derived_param"]
    new_reason = "New derived reason"

    # Act: System updates derived parameter change reason
    derived_param.set_forced_or_derived_change_reason(new_reason)

    # Assert: Derived parameter change reason is updated
    assert derived_param.change_reason == new_reason

    # Arrange: Create regular parameter
    regular_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(param_fixture["param_value"], param_fixture["param_comment"]),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["fc_value"],
    )

    # Act & Assert: System attempting to use method on regular parameter fails
    with pytest.raises(ValueError, match="This method is only for forced or derived parameters"):
        regular_param.set_forced_or_derived_change_reason("Some reason")


def test_forced_or_derived_setters_make_parameter_dirty(param_fixture) -> None:
    """
    Forced/derived setters correctly update dirty state for save detection.

    GIVEN: A forced parameter with matching file values
    WHEN: System updates the value or reason using dedicated setters
    THEN: Parameter should be marked dirty when changed
    AND: Should not be dirty when reset to original values
    """
    # Arrange: Create forced parameter with matching file value and comment
    forced_param = ArduPilotParameter(
        param_fixture["param_name"],
        Par(100.0, "original comment"),
        param_fixture["metadata"],
        param_fixture["default_par"],
        param_fixture["fc_value"],
        forced_par=Par(100.0, "original comment"),
    )

    # Assert: Parameter with matching values is not dirty initially
    assert not forced_param.is_dirty

    # Act: System changes value
    forced_param.set_forced_or_derived_value(200.0)

    # Assert: Parameter is now dirty due to value change
    assert forced_param.is_dirty

    # Act: System resets to original value
    forced_param.set_forced_or_derived_value(100.0)

    # Assert: Parameter is no longer dirty
    assert not forced_param.is_dirty

    # Act: System changes only the reason
    forced_param.set_forced_or_derived_change_reason("new comment")

    # Assert: Parameter is dirty due to comment change
    assert forced_param.is_dirty



"""
Manages configuration steps at the filesystem level.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from json import JSONDecodeError
from json import load as json_load
from logging import error as logging_error
from logging import info as logging_info
from logging import warning as logging_warning
from os import path as os_path
from typing import Optional, TypedDict

# from sys import exit as sys_exit
# from logging import debug as logging_debug
from jsonschema import validate as json_validate
from jsonschema.exceptions import ValidationError

from ardupilot_methodic_configurator import _
from ardupilot_methodic_configurator.data_model_par_dict import Par, ParDict


class PhaseData(TypedDict, total=False):
    """
    Type definition for configuration phase data.

    Attributes:
        start: The starting file number for this phase
        end: The ending file number for this phase (computed)
        weight: The weight for UI layout proportions (computed)
        description: Human-readable description of the phase
        optional: Whether this phase is optional

    """

    start: int
    end: int
    weight: int
    description: str
    optional: bool


class ConfigurationSteps:
    """
    A class to manage configuration steps for the ArduPilot methodic configurator.

    This class provides methods for reading and validating configuration steps, including forced and derived parameters.
    It is designed to simplify the interaction with configuration steps for managing ArduPilot configuration files.

    Attributes:
        configuration_steps_filename (str): The name of the file containing documentation for the configuration files.
        configuration_steps (dict): A dictionary containing the configuration steps.

    """

    def __init__(self, _vehicle_dir: str, vehicle_type: str) -> None:
        self.configuration_steps_filename = "configuration_steps_" + vehicle_type + ".json"
        self.configuration_steps: dict[str, dict] = {}
        self.configuration_phases: dict[str, PhaseData] = {}
        self.forced_parameters: dict[str, ParDict] = {}
        self.derived_parameters: dict[str, ParDict] = {}
        self.log_loaded_file = False

    def re_init(self, vehicle_dir: str, vehicle_type: str) -> None:  # pylint: disable=too-many-branches
        if vehicle_type == "":
            return
        self.configuration_steps_filename = "configuration_steps_" + vehicle_type + ".json"
        # Define a list of directories to search for the configuration_steps_filename file
        search_directories = [vehicle_dir, os_path.dirname(os_path.abspath(__file__))]
        file_found = False
        json_content = {}
        for i, directory in enumerate(search_directories):
            try:
                with open(os_path.join(directory, self.configuration_steps_filename), encoding="utf-8-sig") as file:
                    json_content = json_load(file)
                    file_found = True
                    if self.log_loaded_file:
                        if i == 0:
                            logging_warning(
                                _("Configuration steps '%s' loaded from %s (overwriting default configuration steps)."),
                                self.configuration_steps_filename,
                                directory,
                            )
                        if i == 1:
                            logging_info(
                                _("Configuration steps '%s' loaded from %s."), self.configuration_steps_filename, directory
                            )
                    break
            except FileNotFoundError:
                pass
            except JSONDecodeError as e:
                logging_error(_("Error in file '%s': %s"), self.configuration_steps_filename, e)
                break
        # Validate the vehicle configuration steps file against the configuration_steps_schema.json schema
        if file_found:
            schema_file = os_path.join(os_path.dirname(os_path.abspath(__file__)), "configuration_steps_schema.json")
            try:
                with open(schema_file, encoding="utf-8") as schema:
                    schema_data = json_load(schema)
                    json_validate(instance=json_content, schema=schema_data)
            except FileNotFoundError:
                logging_error(_("Schema file '%s' not found"), schema_file)
            except ValidationError as e:
                logging_error(_("Configuration steps validation error: %s"), str(e))
            except JSONDecodeError as e:
                logging_error(_("Error in schema file '%s': %s"), schema_file, e)

        if file_found and "steps" in json_content:
            self.configuration_steps = json_content["steps"]
            for filename, file_info in self.configuration_steps.items():
                self.__validate_parameters_in_configuration_steps(filename, file_info, "forced")
                self.__validate_parameters_in_configuration_steps(filename, file_info, "derived")
        else:
            logging_warning(_("No configuration steps documentation and no forced and derived parameters will be available."))

        if file_found and "phases" in json_content:
            self.configuration_phases = json_content["phases"]
        else:
            logging_warning(_("No configuration phases documentation will be available."))
        self.log_loaded_file = True

    def __validate_parameters_in_configuration_steps(self, filename: str, file_info: dict, parameter_type: str) -> None:
        """
        Validates the parameters in the configuration steps.

        This method checks if the parameters in the configuration steps are correctly formatted.
        If a parameter is missing the 'New Value' or 'Change Reason' attribute, an error message is logged.
        """
        if parameter_type + "_parameters" in file_info:
            if not isinstance(file_info[parameter_type + "_parameters"], dict):
                logging_error(
                    _("Error in file '%s': '%s' %s parameter is not a dictionary"),
                    self.configuration_steps_filename,
                    filename,
                    parameter_type,
                )
                return
            for parameter, parameter_info in file_info[parameter_type + "_parameters"].items():
                if "New Value" not in parameter_info:
                    logging_error(
                        _("Error in file '%s': '%s' %s parameter '%s' 'New Value' attribute not found."),
                        self.configuration_steps_filename,
                        filename,
                        parameter_type,
                        parameter,
                    )
                if "Change Reason" not in parameter_info:
                    logging_error(
                        _("Error in file '%s': '%s' %s parameter '%s' 'Change Reason' attribute not found."),
                        self.configuration_steps_filename,
                        filename,
                        parameter_type,
                        parameter,
                    )

    def compute_parameters(  # pylint: disable=too-many-branches, too-many-arguments, too-many-positional-arguments
        self,
        filename: str,
        file_info: dict,
        parameter_type: str,
        variables: dict,
        ignore_fc_derived_param_warnings: bool = False,
    ) -> str:
        """
        Computes the forced or derived parameters for a given configuration file.

        If the parameter is forced, it is added to the forced_parameters dictionary.
        If the parameter is derived, it is added to the derived_parameters dictionary.
        """
        if parameter_type + "_parameters" not in file_info or not variables:
            return ""
        destination = self.forced_parameters if parameter_type == "forced" else self.derived_parameters
        for parameter, parameter_info in file_info[parameter_type + "_parameters"].items():  # pylint: disable=too-many-nested-blocks
            try:
                if ("fc_parameters" in str(parameter_info["New Value"])) and (
                    "fc_parameters" not in variables or variables["fc_parameters"] == {}
                ):
                    error_msg = _(
                        "In file '{self.configuration_steps_filename}': '{filename}' {parameter_type} "
                        "parameter '{parameter}' could not be computed: 'fc_parameters' not found, is an FC connected?"
                    )
                    error_msg = error_msg.format(**locals())
                    if parameter_type == "forced":
                        logging_error(error_msg)
                        return error_msg
                    if not ignore_fc_derived_param_warnings:
                        logging_warning(error_msg)
                    continue
                result = eval(str(parameter_info["New Value"]), {}, variables)  # noqa: S307 pylint: disable=eval-used

                # convert (combobox) string text to (parameter value) string int or float
                if isinstance(result, str):
                    if parameter in variables["doc_dict"]:
                        values = variables["doc_dict"][parameter]["values"]
                        if values:
                            result = next(key for key, value in values.items() if value == result)
                        else:
                            bitmasks = variables["doc_dict"][parameter]["Bitmask"]
                            if bitmasks:
                                result = 2 ** next(key for key, bitmask in bitmasks.items() if bitmask == result)
                    else:
                        error_msg = _(
                            "In file '{self.configuration_steps_filename}': '{filename}' {parameter_type} "
                            "parameter '{parameter}' could not be computed, no documentation metadata available for it"
                        )
                        error_msg = error_msg.format(**locals())
                        if parameter_type == "forced":
                            logging_error(error_msg)
                            return error_msg
                        logging_warning(error_msg)
                        continue

                if filename not in destination:
                    destination[filename] = ParDict()
                change_reason = _(parameter_info["Change Reason"]) if parameter_info["Change Reason"] else ""
                destination[filename][parameter] = Par(float(result), change_reason)
            except (SyntaxError, NameError, KeyError, StopIteration) as _e:
                error_msg = _(
                    "In file '{self.configuration_steps_filename}': '{filename}' {parameter_type} "
                    "parameter '{parameter}' could not be computed: {_e}"
                )
                error_msg = error_msg.format(**locals())
                if parameter_type == "forced":
                    logging_error(error_msg)
                    return error_msg
                logging_warning(error_msg)
        return ""

    def auto_changed_by(self, selected_file: str) -> str:
        if selected_file in self.configuration_steps:
            return str(self.configuration_steps[selected_file].get("auto_changed_by", ""))
        return ""

    def jump_possible(self, selected_file: str) -> dict[str, str]:
        if selected_file in self.configuration_steps:
            return dict(self.configuration_steps[selected_file].get("jump_possible", {}))
        return {}

    def get_documentation_text_and_url(self, selected_file: str, prefix_key: str) -> tuple[str, str]:
        documentation = self.configuration_steps.get(selected_file, {}) if self.configuration_steps else None
        if documentation is None:
            text = _(
                "File '{self.configuration_steps_filename}' not found. No intermediate parameter configuration steps available"
            )
            text = text.format(**locals())
            url = ""
        else:
            text = _("No documentation available for {selected_file} in the {self.configuration_steps_filename} file")
            text = documentation.get(prefix_key + "_text", text.format(**locals()))
            url = documentation.get(prefix_key + "_url", "")
        return text, url

    def get_seq_tooltip_text(self, selected_file: str, tooltip_key: str) -> str:
        documentation = self.configuration_steps.get(selected_file, {}) if self.configuration_steps else None
        if documentation is None:
            text = _(
                "File '{self.configuration_steps_filename}' not found. No intermediate parameter configuration steps available"
            )
            text = text.format(**locals())
        else:
            text = _("No documentation available for {selected_file} in the {self.configuration_steps_filename} file")
            text = documentation.get(tooltip_key, text.format(**locals()))
        return text

    def get_sorted_phases_with_end_and_weight(self, total_files: int) -> dict[str, PhaseData]:
        """
        Get sorted phases with added 'end' and 'weight' information.

        Returns phases sorted by start position, with each phase containing:
        - 'end': The end file number (start of next phase or total_files)
        - 'weight': Weight for UI layout (max(2, end - start))
        """
        active_phases = {k: v for k, v in self.configuration_phases.items() if "start" in v}

        # Sort phases by start position
        sorted_phases: dict[str, PhaseData] = dict(sorted(active_phases.items(), key=lambda x: x[1].get("start", 0)))

        # Add the end information to each phase using the start of the next phase
        phase_names = list(sorted_phases.keys())
        for i, phase_name in enumerate(phase_names):
            if i < len(phase_names) - 1:
                next_phase_name = phase_names[i + 1]
                sorted_phases[phase_name]["end"] = sorted_phases[next_phase_name].get("start", total_files)
            else:
                sorted_phases[phase_name]["end"] = total_files
            phase_start = sorted_phases[phase_name].get("start", 0)
            phase_end = sorted_phases[phase_name].get("end", total_files)
            sorted_phases[phase_name]["weight"] = max(2, phase_end - phase_start)

        return sorted_phases

    def get_plugin(self, selected_file: str) -> Optional[dict]:
        """
        Get the plugin configuration for the selected file.

        Args:
            selected_file: The filename to get plugin info for

        Returns:
            The plugin dict with 'name' and 'placement' if exists, None otherwise

        """
        if selected_file in self.configuration_steps:
            return self.configuration_steps[selected_file].get("plugin")
        return None
